import typing

import torch
import triton
from torch import Tensor
from torch._library import triton_op
from torch._library.triton import wrap_triton
from triton import language as tl

from blksprs.utils.autotuning import get_autotune_configs, prune_autotune_configs
from blksprs.utils.blksprs_tensor import BlksprsTensor
from blksprs.utils.tools import stride
from blksprs.utils.validation import validate_dimensions, validate_device, \
    validate_contiguous


@torch.amp.custom_fwd(device_type="cuda", cast_inputs=torch.float16)
def build_distribution_layout(indices: BlksprsTensor, sparsity_layout_indices: Tensor,
                              dim: int, size_target: torch.Size,
                              sparsity_block_size: int) -> Tensor:
    """Builds the sparsity layout of either the source of a gather or the target of a scatter operation.

    Args:
        indices (BlksprsTensor): The block-sparse indices tensor in compressed form used for the gather or scatter operation.
        sparsity_layout_indices (Tensor): The sparsity layout of the indices block-sparse tensor.
        dim (int): The dimension along which the operation is conducted.
        size_target (torch.Size): The size of the block-sparse target tensor in regular form.
        sparsity_block_size (int): The size of the sparsity blocks.

    Returns:
        Tensor: The sparsity layout of the source or target tensor.

    """
    validate_dimensions(indices)
    validate_contiguous(indices)
    validate_device(indices)

    sparsity_lut_i = torch.nonzero(sparsity_layout_indices).contiguous()

    adjusted_dim = dim % 3

    return build_distribution_layout_operation(indices, sparsity_lut_i, adjusted_dim, size_target, sparsity_block_size)


@triton_op("blksprs::build_distribution_layout", mutates_args={})
def build_distribution_layout_operation(indices: Tensor, sparsity_lut_i: Tensor,
                                        adjusted_dim: int, size_target: typing.List[int],
                                        sparsity_block_size: int) -> Tensor:
    with torch.no_grad():
        output = torch.zeros(size_target[0], size_target[1] // sparsity_block_size,
                             size_target[2] // sparsity_block_size,
                             dtype=torch.bool, device=indices.device)

        i_b, i_r, i_c = indices.size()
        i_b_s, i_r_s, i_c_s = stride(indices)
        s_lut_i_r, s_lut_i_c = sparsity_lut_i.size()
        s_lut_i_r_s, s_lut_i_c_s = stride(sparsity_lut_i)
        o_b, o_r, o_c = output.size()
        o_b_s, o_r_s, o_c_s = stride(output)

        triton_grid = lambda meta: [i_b,
                                    triton.cdiv(i_r, meta["TRITON_BLOCK_SIZE"]),
                                    triton.cdiv(i_c, meta["TRITON_BLOCK_SIZE"])]

        (wrap_triton(build_distribution_layout_kernel)[triton_grid]
         (indices,
          i_b, i_b_s, i_r_s, i_c_s,
          sparsity_lut_i,
          s_lut_i_r, s_lut_i_r_s, s_lut_i_c_s,
          adjusted_dim,
          output,
          o_b, o_b_s, o_r_s, o_c_s,
          sparsity_block_size))

        return output


@triton.autotune(
    configs=get_autotune_configs(),
    key=["sparsity_block_size"],
    prune_configs_by={"early_config_prune": prune_autotune_configs},
    reset_to_zero=["o"]
)
@triton.jit
def build_distribution_layout_kernel(i,
                                     i_b, i_b_s, i_r_s, i_c_s,
                                     s_lut_i,
                                     s_lut_i_r, s_lut_i_r_s, s_lut_i_c_s,
                                     dim,
                                     o,
                                     o_b, o_b_s, o_r_s, o_c_s,
                                     sparsity_block_size,
                                     TRITON_BLOCK_SIZE: tl.constexpr) -> None:
    # Get triton block indices
    pid_blk = tl.program_id(axis=0)
    pid_row = tl.program_id(axis=1)
    pid_col = tl.program_id(axis=2)

    # Get position of current sparsity block consisting of its batch, row, and column index
    spa_bat_i_idx = (pid_blk * s_lut_i_r_s + 0 * s_lut_i_c_s)
    spa_bat_i_msk = ((spa_bat_i_idx >= 0) &
                     (spa_bat_i_idx < s_lut_i_r * s_lut_i_r_s))
    spa_bat_i = tl.load(s_lut_i + spa_bat_i_idx, mask=spa_bat_i_msk)

    spa_row_i_idx = (pid_blk * s_lut_i_r_s + 1 * s_lut_i_c_s)
    spa_row_i_msk = ((spa_row_i_idx >= 0) &
                     (spa_row_i_idx < s_lut_i_r * s_lut_i_r_s))
    spa_row_i = tl.load(s_lut_i + spa_row_i_idx, mask=spa_row_i_msk)

    spa_col_i_idx = (pid_blk * s_lut_i_r_s + 2 * s_lut_i_c_s)
    spa_col_i_msk = ((spa_col_i_idx >= 0) &
                     (spa_col_i_idx < s_lut_i_r * s_lut_i_r_s))
    spa_col_i = tl.load(s_lut_i + spa_col_i_idx, mask=spa_col_i_msk)

    blk_i_idx = (pid_blk * i_b_s +
                 ((pid_row * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * i_r_s)[:, None] +
                 ((pid_col * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * i_c_s)[None, :])
    blk_i_msk = ((blk_i_idx >= 0) &
                 (blk_i_idx < i_b * i_b_s))
    blk_i = tl.load(i + blk_i_idx, mask=blk_i_msk)

    dst_bat_idx = tl.full((TRITON_BLOCK_SIZE, TRITON_BLOCK_SIZE), spa_bat_i, dtype=tl.int32)
    dst_row_idx = tl.full((TRITON_BLOCK_SIZE, TRITON_BLOCK_SIZE), spa_row_i, dtype=tl.int32)
    dst_col_idx = tl.full((TRITON_BLOCK_SIZE, TRITON_BLOCK_SIZE), spa_col_i, dtype=tl.int32)
    if dim == 0:
        dst_bat_idx = blk_i
    elif dim == 1:
        dst_row_idx = blk_i // sparsity_block_size
    elif dim == 2:
        dst_col_idx = blk_i // sparsity_block_size

    blk_v = tl.full((TRITON_BLOCK_SIZE, TRITON_BLOCK_SIZE), 1, dtype=tl.int1)

    blk_o_idx = ((dst_bat_idx * o_b_s) +
                 (dst_row_idx * o_r_s) +
                 (dst_col_idx * o_c_s))
    blk_o_msk = ((blk_o_idx >= 0) &
                 (blk_o_idx < o_b * o_b_s))
    tl.store(o + blk_o_idx, blk_v, mask=blk_o_msk)



import torch
from torch import Tensor
from torch._library import triton_op

from blksprs.ops.flow import flow_pull_forward
from blksprs.utils.blksprs_tensor import BlksprsTensor
from blksprs.utils.validation import validate_dimensions, validate_contiguous, validate_device, \
    validate_sparsity, validate_sparsity_block_size, ensure_contiguous


@torch.amp.custom_fwd(device_type="cuda", cast_inputs=torch.float16)
def transpose(x: BlksprsTensor, sparsity_layout: Tensor,
              sparsity_block_size: int, lut: dict = None) -> (BlksprsTensor, Tensor):
    """Transposes a block-sparse tensor in compressed form.

    Note:
         Returns the transposed tensor and the sparsity layout of the transposed tensor.

    Args:
        x (BlksprsTensor): A block-sparse tensor in compressed form.
        sparsity_layout (Tensor): The sparsity layout of the block-sparse tensor.
        sparsity_block_size (int): The size of the sparsity blocks.
        lut (dict, optional): A dictionary containing the look-up tables for the operation (default ``None``).

    Returns:
        BlksprsTensor: The transposed block-sparse tensor in compressed form.
        Tensor: The sparsity layout of the transposed tensor.

    """
    x = ensure_contiguous(x)

    validate_dimensions(x)
    validate_contiguous(x)
    validate_device(x)
    validate_sparsity(sparsity_block_size, (x, sparsity_layout))
    validate_sparsity_block_size(sparsity_block_size, x)

    lut = transpose_build_lut(lut, sparsity_layout)

    return BlksprsTensor.wrap(transpose_forward(x, lut["sparsity_layout_t"],
                                                lut["sparsity_lut"], lut["sparsity_reverse_lut"],
                                                sparsity_block_size, lut["n_sparse_blocks"])), lut["sparsity_layout_t"]


@triton_op("blksprs::transpose_forward", mutates_args={})
def transpose_forward(x: Tensor, sparsity_layout_o: Tensor,
                      sparsity_lut: Tensor, sparsity_reverse_lut: Tensor,
                      sparsity_block_size: int, n_sparse_blocks: int) -> Tensor:
    with torch.no_grad():
        x_t = x.transpose(-1, -2).contiguous()
        return flow_pull_forward(x_t, sparsity_layout_o, sparsity_lut, sparsity_reverse_lut,
                                 sparsity_block_size, n_sparse_blocks)


def transpose_wrapper_backward(ctx, grad_output):
    sparsity_layout = ctx.saved_tensors[0]
    sparsity_block_size = ctx.sparsity_block_size

    return transpose(grad_output, sparsity_layout, sparsity_block_size)[
        0], None, None, None, None, None


def transpose_build_lut(lut: dict, sparsity_layout: Tensor):
    if lut is None:
        lut = dict()

    if "sparsity_layout_t" not in lut:
        sparsity_layout_t = sparsity_layout.transpose(-1, -2).contiguous()
        lut["sparsity_layout_t"] = sparsity_layout_t

    if "sparsity_lut" not in lut:
        sparsity_lut = torch.nonzero(lut["sparsity_layout_t"]).contiguous()
        lut["sparsity_lut"] = sparsity_lut

    if "sparsity_reverse_lut" not in lut:
        sparsity_layout_flat = sparsity_layout.reshape(-1)
        sparsity_reverse_lut = (((torch.cumsum(sparsity_layout_flat, dim=-1) - 1) *
                                 (sparsity_layout_flat == 1) -
                                 (1 * (sparsity_layout_flat == 0)))
                                .reshape(sparsity_layout.size()).transpose(-1, -2).contiguous().reshape(-1))
        lut["sparsity_reverse_lut"] = sparsity_reverse_lut

    if "n_sparse_blocks" not in lut:
        n_sparse_blocks = torch.sum(sparsity_layout.to(torch.int)).item()
        lut["n_sparse_blocks"] = n_sparse_blocks

    validate_contiguous(lut["sparsity_layout_t"], lut["sparsity_lut"], lut["sparsity_reverse_lut"])

    return lut


# noinspection PyUnusedLocal
def transpose_setup_context(ctx, inputs, output):
    (_, sparsity_layout_o, _, _, sparsity_block_size, _) = inputs

    ctx.save_for_backward(sparsity_layout_o)
    ctx.sparsity_block_size = sparsity_block_size


transpose_forward.register_autograd(transpose_wrapper_backward, setup_context=transpose_setup_context)



# Settings
import torch

# Capture scalar outputs for JIT compilation
torch._dynamo.config.capture_scalar_outputs = True
# Set version
__version__ = "2.2"

# Imports

from blksprs.utils.blksprs_tensor import BlksprsTensor


class ops:
    from blksprs.ops.conversion import to_dense, to_sparse, from_blksprs, to_blksprs, adapt_layout
    from blksprs.ops.distribution import gather, scatter, scatter_reduce
    from blksprs.ops.flash_attention import flash_attention, flash_attention_build_lut
    from blksprs.ops.matmul import matmul
    from blksprs.ops.softmax import softmax, softmax_fused
    from blksprs.ops.transpose import transpose
    from blksprs.ops.repeat import repeat, repeat_interleave
    from blksprs.ops.partitioning import split, merge

    class misc:
        from blksprs.ops.misc.row_wise import row_wise_sum, row_wise_max, row_wise_add, row_wise_sub
        from blksprs.ops.misc.broadcast_ops import broadcast_add, broadcast_sub


class layouting:
    from blksprs.layouting.distribution_layout import build_distribution_layout
    from blksprs.layouting.sparsity_layout import build_sparsity_layout, build_sparsity_layout_adaption, \
        build_sparsity_layout_matmul, build_sparsity_layout_matmul_fast, build_sparsity_layout_full


class utils:
    from blksprs.utils.processing import apply_torch_linear, apply_torch_normalisation, apply_torch_dropout, \
        apply_function_applicable_row_wise
    from blksprs.utils.tools import do_shape_blocksparse, undo_shape_blocksparse
    from blksprs.utils.validation import disable_contiguous, disable_validation

    class validation:
        from blksprs.utils.validation import validate_dimensions, validate_contiguous, validate_dtype_float, \
            validate_dtype_int, validate_device, validate_sparsity, validate_sparsity_dense, \
            validate_sparsity_block_size



import os

blksprs_autotune_mode = os.getenv("BLKSPRS_AUTOTUNE", "DEFAULT")

if blksprs_autotune_mode == "DEFAULT":
    autotune_parameters = [
        (16, 3, 8),
        (16, 4, 4),
        (16, 5, 2),

        (32, 3, 8),
        (32, 4, 4),
        (32, 5, 2),

        (64, 3, 8),
        (64, 4, 4),
        (64, 4, 8),

        (128, 3, 8),
        (128, 4, 4),
        (128, 4, 8),
    ]
elif blksprs_autotune_mode == "TEST":
    autotune_parameters = [
        (16, 3, 8),

        (32, 3, 8),

        (64, 3, 8),
    ]
else:
    raise NotImplementedError(f"Unknown autotune mode: {blksprs_autotune_mode}")

import torch
import triton


def prune_autotune_configs(autotune_configs, kernel_args, **kwargs):
    sparsity_block_size = kernel_args["sparsity_block_size"]

    pruned_configs = []

    for config in autotune_configs:
        if config.kwargs["TRITON_BLOCK_SIZE"] <= sparsity_block_size:
            pruned_configs.append(config)

    assert len(pruned_configs) > 0, f"No valid autotune configs found for sparsity block size {sparsity_block_size}"

    return pruned_configs


def prune_autotune_configs_conversion(autotune_configs, kernel_args, **kwargs):
    sparsity_block_size_from = kernel_args["sparsity_block_size_from"]
    sparsity_block_size_to = kernel_args["sparsity_block_size_to"]
    sparsity_block_size = min(sparsity_block_size_from, sparsity_block_size_to)

    pruned_configs = []

    for config in autotune_configs:
        if config.kwargs["TRITON_BLOCK_SIZE"] <= sparsity_block_size:
            pruned_configs.append(config)

    assert len(pruned_configs) > 0, f"No valid autotune configs found for sparsity block size {sparsity_block_size}"

    return pruned_configs


def get_autotune_configs():
    global autotune_parameters

    autotune_configs = []

    for block_size, num_stages, num_warps in autotune_parameters:
        autotune_configs.append(
            triton.Config({"TRITON_BLOCK_SIZE": block_size}, num_stages=num_stages, num_warps=num_warps))

    return autotune_configs



import torch
import triton
from torch import Tensor
from torch._library import triton_op
from torch._library.triton import wrap_triton
from triton import language as tl

from blksprs.ops.misc.row_wise import row_wise_sum, row_wise_max, row_wise_sub
from blksprs.utils.autotuning import get_autotune_configs, prune_autotune_configs
from blksprs.utils.blksprs_tensor import BlksprsTensor
from blksprs.utils.tools import stride, ceil_pow2
from blksprs.utils.validation import validate_contiguous, validate_dimensions, validate_device, \
    validate_sparsity, validate_sparsity_block_size, validate_dtype_float_32, ensure_contiguous


def softmax(x: BlksprsTensor, sparsity_layout: Tensor, sparsity_block_size: int, flag_fused: bool = True,
            lut: dict = None) -> BlksprsTensor:
    """Wrapper for :func:`softmax_regular` and :func:`softmax_fused` based on the ``flag_fused`` parameter.

    """
    if flag_fused:
        return softmax_fused(x, sparsity_layout, sparsity_block_size, lut)
    else:
        return softmax_regular(x, sparsity_layout, sparsity_block_size, lut)


@torch.amp.custom_fwd(device_type="cuda", cast_inputs=torch.float32)
def softmax_regular(x: BlksprsTensor, sparsity_layout: Tensor, sparsity_block_size: int,
                    lut: dict = None) -> BlksprsTensor:
    """Computes the softmax of a block-sparse tensor in compressed form.

    Note:
        Sparse blocks are not considered for the calculation of the softmax, i.e., all values are assumed to be ``-inf``.

    Args:
        x (BlksprsTensor): A block-sparse tensor in compressed form.
        sparsity_layout (Tensor): The sparsity layout of the block-sparse tensor.
        sparsity_block_size (int): The size of the sparsity blocks.
        lut (dict, optional): A dictionary containing the look-up tables for the operation (default ``None``).

    Returns:
        BlksprsTensor: The result of the softmax operation as a block-sparse tensor in compressed form.

    """
    x = ensure_contiguous(x)

    validate_dimensions(x)
    validate_contiguous(x)
    validate_dtype_float_32(x)
    validate_device(x)
    validate_sparsity(sparsity_block_size, (x, sparsity_layout))
    validate_sparsity_block_size(sparsity_block_size, x)

    lut = softmax_build_lut(lut, sparsity_layout)

    return BlksprsTensor.wrap(softmax_forward(x, sparsity_layout,
                                              lut["sparsity_lut"],
                                              lut["sparsity_reverse_lut_rws"],
                                              sparsity_block_size))


@triton_op("blksprs::softmax_forward", mutates_args={})
def softmax_forward(x: Tensor, sparsity_layout: Tensor,
                    sparsity_lut: Tensor,
                    sparsity_reverse_lut_rws: Tensor,
                    sparsity_block_size: int) -> Tensor:
    output = torch.empty_like(x)

    x_row_wise_max, sparsity_layout_rwm = row_wise_max(x, sparsity_layout, sparsity_block_size,
                                                       flag_slice_only=True)
    x_scaled = row_wise_sub(x, sparsity_layout, x_row_wise_max, sparsity_block_size)
    x_exp = torch.exp(x_scaled)
    x_exp_row_wise_sum, sparsity_layout_rws = row_wise_sum(x_exp, sparsity_layout, sparsity_block_size,
                                                           flag_slice_only=True)

    x_b, x_r, x_c = x.size()
    x_b_s, x_r_s, x_c_s = stride(x)
    s_lut_r, s_lut_c = sparsity_lut.size()
    s_lut_r_s, s_lut_c_s = stride(sparsity_lut)
    o_b, o_r, o_c = output.size()
    s_b, s_r, s_c = x_exp_row_wise_sum.shape
    s_b_s, s_r_s, s_c_s = stride(x_exp_row_wise_sum)
    s_l_s_b, s_l_s_r, s_l_s_c = sparsity_layout_rws.shape
    s_l_s_b_s, s_l_s_r_s, s_l_s_c_s = stride(sparsity_layout_rws)

    triton_grid = lambda meta: [o_b,
                                triton.cdiv(o_r, meta["TRITON_BLOCK_SIZE"]),
                                triton.cdiv(o_c, meta["TRITON_BLOCK_SIZE"])]

    (wrap_triton(softmax_kernel)[triton_grid]
     (x_exp,
      x_b, x_b_s, x_r_s, x_c_s,
      sparsity_lut, s_lut_r, s_lut_r_s, s_lut_c_s,
      x_exp_row_wise_sum, s_b, s_b_s, s_r_s, s_c_s,
      s_l_s_b, s_l_s_b_s, s_l_s_r_s,
      sparsity_reverse_lut_rws,
      output,
      sparsity_block_size))

    return output


def softmax_backward_wrapper(ctx, grad_output):
    o, sparsity_layout, sparsity_lut = ctx.saved_tensors
    sparsity_block_size = ctx.sparsity_block_size

    return softmax_backward(grad_output, o, sparsity_lut, sparsity_layout,
                            sparsity_block_size), None, None, None, None, None


@triton_op("blksprs::softmax_backward", mutates_args={})
def softmax_backward(grad_output: Tensor, o: Tensor, sparsity_lut: Tensor, sparsity_layout: Tensor,
                     sparsity_block_size: int) -> Tensor:
    with torch.no_grad():
        grad_x = torch.empty_like(o, dtype=torch.float)

        s, sparsity_layout_s = row_wise_sum(grad_output * o, sparsity_layout, sparsity_block_size, flag_slice_only=True)

        sparsity_layout_s_flat = sparsity_layout_s.reshape(-1)
        sparsity_reverse_lut_s = ((torch.cumsum(sparsity_layout_s_flat, dim=-1) - 1) *
                                  (sparsity_layout_s_flat == 1) -
                                  (1 * (sparsity_layout_s_flat == 0)))

        o_b, o_r, o_c = o.size()
        o_b_s, o_r_s, o_c_s = stride(o)
        s_lut_r, s_lut_c = sparsity_lut.size()
        s_lut_r_s, s_lut_c_s = stride(sparsity_lut)
        s_b, s_r, s_c = s.size()
        s_b_s, s_r_s, s_c_s = stride(s)
        s_l_s_b, s_l_s_r, s_l_s_c = sparsity_layout_s.size()
        s_l_s_b_s, s_l_s_r_s, s_l_s_c_s = stride(sparsity_layout_s)

        triton_grid = lambda meta: [o_b,
                                    triton.cdiv(o_r, meta["TRITON_BLOCK_SIZE"]),
                                    triton.cdiv(o_c, meta["TRITON_BLOCK_SIZE"])]

        (wrap_triton(softmax_kernel_grad)[triton_grid]
         (grad_output,
          o_b, o_b_s, o_r_s, o_c_s,
          o,
          o_b, o_b_s, o_r_s, o_c_s,
          sparsity_lut, s_lut_r, s_lut_r_s, s_lut_c_s,
          s,
          s_b, s_b_s, s_r_s, s_c_s,
          s_l_s_b, s_l_s_b_s, s_l_s_r_s,
          sparsity_reverse_lut_s,
          grad_x,
          o_b, o_b_s, o_r_s, o_c_s,
          sparsity_block_size))

        return grad_x


# noinspection PyUnusedLocal
@triton.autotune(
    configs=get_autotune_configs(),
    key=["sparsity_block_size"],
    prune_configs_by={"early_config_prune": prune_autotune_configs},
    reset_to_zero=["o"]
)
@triton.jit
def softmax_kernel(x,
                   x_b, x_b_s, x_r_s, x_c_s,
                   s_lut, s_lut_r, s_lut_r_s, s_lut_c_s,
                   s, s_b, s_b_s, s_r_s, s_c_s,
                   s_l_s_b, s_l_s_b_s, s_l_s_r_s,
                   r_lut_s,
                   o,
                   sparsity_block_size,
                   TRITON_BLOCK_SIZE: tl.constexpr) -> None:
    # Get triton block indices
    pid_blk = tl.program_id(axis=0)
    pid_row = tl.program_id(axis=1)
    pid_col = tl.program_id(axis=2)

    # Get position of current sparsity block consisting of its batch and row index
    spa_val_idx = pid_blk * s_lut_r_s + tl.arange(0, 4) * s_lut_c_s
    spa_val_msk = (tl.arange(0, 4) < 3)
    spa_val = tl.load(s_lut + spa_val_idx, mask=spa_val_msk)

    spa_bat = tl.sum(spa_val * (tl.arange(0, 4) == 0))
    spa_row = tl.sum(spa_val * (tl.arange(0, 4) == 1))

    # Get reverse sparsity indices for s
    rev_idx_spa_s_idx = (spa_bat * s_l_s_b_s +
                         spa_row * s_l_s_r_s)
    rev_idx_spa_s_msk = ((rev_idx_spa_s_idx >= 0) &
                         (rev_idx_spa_s_idx < s_l_s_b * s_l_s_b_s))
    rev_idx_spa_s = tl.load(r_lut_s + rev_idx_spa_s_idx, mask=rev_idx_spa_s_msk).to(tl.int32)

    if rev_idx_spa_s >= 0:
        # Load x block
        blk_x_idx = ((pid_blk * x_b_s) +
                     ((pid_row * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * x_r_s)[:, None] +
                     ((pid_col * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * x_c_s)[None, :])
        blk_x_msk = ((blk_x_idx >= 0) &
                     (blk_x_idx < x_b * x_b_s))
        blk_x = tl.load(x + blk_x_idx, mask=blk_x_msk)

        # Load sum block
        blk_s_idx = (rev_idx_spa_s * s_b_s +
                     ((pid_row * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * s_r_s)[:, None] +
                     (tl.arange(0, 1) * s_c_s)[None, :])
        blk_s_msk = ((blk_s_idx >= 0) &
                     (blk_s_idx < s_b * s_b_s))
        blk_s = tl.load(s + blk_s_idx, mask=blk_s_msk)

        # Compute softmax
        buf = tl.div_rn(blk_x, blk_s)

        # Store output
        tl.store(o + blk_x_idx, buf, mask=blk_x_msk)


# noinspection PyUnusedLocal
@triton.autotune(
    configs=get_autotune_configs(),
    key=["sparsity_block_size"],
    prune_configs_by={"early_config_prune": prune_autotune_configs},
    reset_to_zero=["o"]
)
@triton.jit
def softmax_kernel_grad(g,
                        g_b, g_b_s, g_r_s, g_c_s,
                        x,
                        x_b, x_b_s, x_r_s, x_c_s,
                        s_lut, s_lut_r, s_lut_r_s, s_lut_c_s,
                        s,
                        s_b, s_b_s, s_r_s, s_c_s,
                        s_l_s_b, s_l_s_b_s, s_l_s_r_s,
                        r_lut_s,
                        o,
                        o_b, o_b_s, o_r_s, o_c_s,
                        sparsity_block_size,
                        TRITON_BLOCK_SIZE: tl.constexpr) -> None:
    # Get triton block indices
    pid_blk = tl.program_id(axis=0)
    pid_row = tl.program_id(axis=1)
    pid_col = tl.program_id(axis=2)

    # Get position of current sparsity block consisting of its batch and row index
    spa_val_idx = pid_blk * s_lut_r_s + tl.arange(0, 4) * s_lut_c_s
    spa_val_msk = (tl.arange(0, 4) < 3)
    spa_val = tl.load(s_lut + spa_val_idx, mask=spa_val_msk)

    spa_bat = tl.sum(spa_val * (tl.arange(0, 4) == 0))
    spa_row = tl.sum(spa_val * (tl.arange(0, 4) == 1))

    rev_idx_spa_s_idx = (spa_bat * s_l_s_b_s +
                         spa_row * s_l_s_r_s)
    rev_idx_spa_s_msk = ((rev_idx_spa_s_idx >= 0) &
                         (rev_idx_spa_s_idx < s_l_s_b * s_l_s_b_s))
    rev_idx_spa_s = tl.load(r_lut_s + rev_idx_spa_s_idx, mask=rev_idx_spa_s_msk).to(tl.int32)

    if rev_idx_spa_s >= 0:
        blk_s_idx = (rev_idx_spa_s * s_b_s +
                     ((pid_row * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * s_r_s)[:, None] +
                     (tl.arange(0, 1) * s_c_s)[None, :])
        blk_s_msk = ((blk_s_idx >= 0) &
                     (blk_s_idx < s_b * s_b_s))
        blk_s = tl.load(s + blk_s_idx, mask=blk_s_msk)

        blk_g_idx = ((pid_blk * g_b_s) +
                     ((pid_row * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * g_r_s)[:, None] +
                     ((pid_col * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * g_c_s)[None, :])
        blk_g_msk = ((blk_g_idx >= 0) &
                     (blk_g_idx < g_b * g_b_s))
        blk_g = tl.load(g + blk_g_idx, mask=blk_g_msk)

        blk_x_idx = ((pid_blk * x_b_s) +
                     ((pid_row * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * x_r_s)[:, None] +
                     ((pid_col * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * x_c_s)[None, :])
        blk_x_msk = ((blk_x_idx >= 0) &
                     (blk_x_idx < x_b * x_b_s))
        blk_x = tl.load(x + blk_x_idx, mask=blk_x_msk)

        buf = blk_x * (blk_g - blk_s)

        blk_o_idx = ((pid_blk * o_b_s) +
                     ((pid_row * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * o_r_s)[:, None] +
                     ((pid_col * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * o_c_s)[None, :])
        blk_o_msk = ((blk_o_idx >= 0) &
                     (blk_o_idx < o_b * o_b_s))
        tl.store(o + blk_o_idx, buf, mask=blk_o_msk)


def softmax_build_lut(lut: dict, sparsity_layout: Tensor):
    if lut is None:
        lut = dict()

    if "sparsity_lut" not in lut:
        sparsity_lut = torch.nonzero(sparsity_layout).contiguous()
        lut["sparsity_lut"] = sparsity_lut

    if "sparsity_reverse_lut_rws" not in lut:
        sparsity_layout_rws, _ = torch.max(sparsity_layout, dim=-1, keepdim=True)
        sparsity_layout_rws_flat = sparsity_layout_rws.reshape(-1)
        sparsity_reverse_lut_rws = ((torch.cumsum(sparsity_layout_rws_flat, dim=-1) - 1) *
                                    (sparsity_layout_rws_flat == 1) -
                                    (1 * (sparsity_layout_rws_flat == 0)))
        lut["sparsity_reverse_lut_rws"] = sparsity_reverse_lut_rws

    validate_contiguous(sparsity_layout, lut["sparsity_lut"], lut["sparsity_reverse_lut_rws"])

    return lut


# noinspection PyUnusedLocal
def softmax_setup_context(ctx, inputs, output):
    (_, sparsity_layout, sparsity_lut, _, sparsity_block_size) = inputs

    ctx.save_for_backward(output, sparsity_layout, sparsity_lut)
    ctx.sparsity_block_size = sparsity_block_size


softmax_forward.register_autograd(softmax_backward_wrapper, setup_context=softmax_setup_context)


@torch.amp.custom_fwd(device_type="cuda", cast_inputs=torch.float32)
def softmax_fused(x: BlksprsTensor, sparsity_layout: Tensor, sparsity_block_size: int,
                  lut: dict = None) -> BlksprsTensor:
    """Computes the softmax fused for each row of a block-sparse tensor in compressed form.

    Note:
        This softmax implementation is a fused version that loads the entire row of a block-sparse tensor into memory.
        See :func:`softmax` for a true block-wise softmax implementation.

    Args:
        x (BlksprsTensor): A block-sparse tensor in compressed form.
        sparsity_layout (Tensor): The sparsity layout of the block-sparse tensor.
        sparsity_block_size (int): The size of the sparsity blocks.
        lut (dict, optional): A dictionary containing the look-up tables for the operation (default ``None``).

    Returns:
        BlksprsTensor: The result of the softmax operation as a block-sparse tensor in compressed form.

    """
    x = ensure_contiguous(x)

    validate_dimensions(x)
    validate_contiguous(x)
    validate_dtype_float_32(x)
    validate_device(x)
    validate_sparsity(sparsity_block_size, (x, sparsity_layout))
    validate_sparsity_block_size(sparsity_block_size, x)

    lut = softmax_fused_build_lut(lut, sparsity_layout)

    return BlksprsTensor.wrap(softmax_fused_forward(x, sparsity_layout,
                                                    lut["sparsity_reverse_lut_sorted"],
                                                    lut["max_blocks_line"],
                                                    sparsity_block_size))


@triton_op("blksprs::softmax_fused_forward", mutates_args={})
def softmax_fused_forward(x: Tensor, sparsity_layout: Tensor,
                          sparsity_reverse_lut_sorted: Tensor,
                          max_blocks_line: int,
                          sparsity_block_size: int) -> Tensor:
    output = torch.empty_like(x)

    x_b, x_r, x_c = x.size()
    x_b_s, x_r_s, x_c_s = stride(x)
    s_l_b, s_l_r, s_l_c = sparsity_layout.size()
    s_l_b_s, s_l_r_s, s_l_c_s = stride(sparsity_layout)

    triton_grid = lambda meta: [s_l_b,
                                s_l_r,
                                sparsity_block_size]

    (wrap_triton(softmax_fused_kernel)[triton_grid]
     (x,
      x_b, x_b_s, x_r_s, x_c_s,
      output,
      s_l_b, s_l_b_s, s_l_r_s, s_l_c, s_l_c_s,
      sparsity_reverse_lut_sorted,
      max_blocks_line,
      sparsity_block_size))

    return output


def softmax_fused_backward_wrapper(ctx, grad_output):
    o, sparsity_layout, sparsity_reverse_lut_sorted = ctx.saved_tensors
    max_blocks_line = ctx.max_blocks_line
    sparsity_block_size = ctx.sparsity_block_size

    return softmax_fused_backward(grad_output, o, sparsity_reverse_lut_sorted, sparsity_layout,
                                  max_blocks_line, sparsity_block_size), None, None, None, None


@triton_op("blksprs::softmax_fused_backward", mutates_args={})
def softmax_fused_backward(grad_output: Tensor,
                           o: Tensor,
                           sparsity_reverse_lut_sorted: Tensor,
                           sparsity_layout: Tensor,
                           max_blocks_line: int,
                           sparsity_block_size: int) -> Tensor:
    with torch.no_grad():
        grad_x = torch.empty_like(o)

        g_b, g_r, g_c = grad_output.size()
        g_b_s, g_r_s, g_c_s = stride(grad_output)
        o_b, o_r, o_c = o.size()
        o_b_s, o_r_s, o_c_s = stride(o)
        s_l_b, s_l_r, s_l_c = sparsity_layout.size()
        s_l_b_s, s_l_r_s, s_l_c_s = stride(sparsity_layout)

        triton_grid = lambda meta: [s_l_b,
                                    s_l_r,
                                    sparsity_block_size]

        (wrap_triton(softmax_fused_kernel_grad)[triton_grid]
         (grad_output,
          g_b, g_b_s, g_r_s, g_c_s,
          o,
          o_b, o_b_s, o_r_s, o_c_s,
          s_l_b, s_l_b_s, s_l_r_s, s_l_c, s_l_c_s,
          sparsity_reverse_lut_sorted,
          grad_x,
          max_blocks_line,
          sparsity_block_size))

        return grad_x


# noinspection PyUnusedLocal
@triton.autotune(
    configs=get_autotune_configs(),
    key=["sparsity_block_size"],
    prune_configs_by={"early_config_prune": prune_autotune_configs},
    reset_to_zero=["o"]
)
@triton.jit
def softmax_fused_kernel(x,
                         x_b, x_b_s, x_r_s, x_c_s,
                         o,
                         s_l_b, s_l_b_s, s_l_r_s, s_l_c, s_l_c_s,
                         r_lut_s,
                         mbs: tl.constexpr,
                         sparsity_block_size: tl.constexpr,
                         TRITON_BLOCK_SIZE: tl.constexpr) -> None:
    # Get triton block indices
    pid_bat = tl.program_id(axis=0)
    pid_row = tl.program_id(axis=1)
    pid_lin = tl.program_id(axis=2)

    # Load reverse sparsity indices of row
    blk_rev_idx = (pid_bat * s_l_b_s +
                   pid_row * s_l_r_s +
                   (tl.arange(0, mbs) * s_l_c_s))
    blk_rev_msk = (((blk_rev_idx >= 0) &
                    (blk_rev_idx < s_l_b * s_l_b_s)) &
                   (tl.arange(0, mbs) < s_l_c))
    blk_rev = tl.load(r_lut_s + blk_rev_idx, mask=blk_rev_msk, other=-1).to(tl.int32)

    if (not (tl.min(blk_rev) == -1 and
             tl.max(blk_rev) == -1)):
        # Extend sparsity indices to cover sparsity blocks
        blk_rev_ext = tl.expand_dims(blk_rev, -1)
        blk_rev_ext = tl.broadcast_to(blk_rev_ext, (mbs, sparsity_block_size))
        blk_rev_ext = tl.reshape(blk_rev_ext, (mbs * sparsity_block_size))

        # Load line of x
        blk_x_idx = (blk_rev_ext * x_b_s +
                     pid_lin * x_r_s +
                     (tl.arange(0, mbs * sparsity_block_size) % sparsity_block_size) * x_c_s)
        blk_x_mask = (((blk_x_idx >= 0) &
                       (blk_x_idx < x_b * x_b_s)) &
                      (blk_rev_ext != -1))
        blk_x = tl.load(x + blk_x_idx, mask=blk_x_mask, other=float("-inf"))

        # Compute softmax
        blk_x_softmax = tl.softmax(blk_x)

        # Store output
        tl.store(o + blk_x_idx, blk_x_softmax, mask=blk_x_mask)


# noinspection PyUnusedLocal
@triton.autotune(
    configs=get_autotune_configs(),
    key=["sparsity_block_size"],
    prune_configs_by={"early_config_prune": prune_autotune_configs},
    reset_to_zero=["o"]
)
@triton.jit
def softmax_fused_kernel_grad(g,
                              g_b, g_b_s, g_r_s, g_c_s,
                              x,
                              x_b, x_b_s, x_r_s, x_c_s,
                              s_l_b, s_l_b_s, s_l_r_s, s_l_c, s_l_c_s,
                              r_lut_s,
                              o,
                              mbs: tl.constexpr,
                              sparsity_block_size: tl.constexpr,
                              TRITON_BLOCK_SIZE: tl.constexpr) -> None:
    # Get triton block indices
    pid_bat = tl.program_id(axis=0)
    pid_row = tl.program_id(axis=1)
    pid_lin = tl.program_id(axis=2)

    # Load reverse sparsity indices of row
    blk_rev_idx = (pid_bat * s_l_b_s +
                   pid_row * s_l_r_s +
                   (tl.arange(0, mbs) * s_l_c_s))
    blk_rev_msk = (((blk_rev_idx >= 0) &
                    (blk_rev_idx < s_l_b * s_l_b_s)) &
                   (tl.arange(0, mbs) < s_l_c))
    blk_rev = tl.load(r_lut_s + blk_rev_idx, mask=blk_rev_msk, other=-1).to(tl.int32)

    if (not (tl.min(blk_rev) == -1 and
             tl.max(blk_rev) == -1)):
        # Extend sparsity indices to cover sparsity blocks
        blk_rev_ext = tl.expand_dims(blk_rev, -1)
        blk_rev_ext = tl.broadcast_to(blk_rev_ext, (mbs, sparsity_block_size))
        blk_rev_ext = tl.reshape(blk_rev_ext, (mbs * sparsity_block_size))

        # Load line of g
        blk_g_idx = (blk_rev_ext * g_b_s +
                     pid_lin * g_r_s +
                     (tl.arange(0, mbs * sparsity_block_size) % sparsity_block_size) * g_c_s)
        blk_g_mask = (((blk_g_idx >= 0) &
                       (blk_g_idx < g_b * g_b_s)) &
                      (blk_rev_ext != -1))
        blk_g = tl.load(g + blk_g_idx, mask=blk_g_mask)

        # Load line of x
        blk_x_idx = (blk_rev_ext * x_b_s +
                     pid_lin * x_r_s +
                     (tl.arange(0, mbs * sparsity_block_size) % sparsity_block_size) * x_c_s)
        blk_x_mask = (((blk_x_idx >= 0) &
                       (blk_x_idx < x_b * x_b_s)) &
                      (blk_rev_ext != -1))
        blk_x = tl.load(x + blk_x_idx, mask=blk_x_mask)

        # Compute gradients
        blk_grad = blk_x * (blk_g - tl.sum(blk_x * blk_g))

        # Store output
        tl.store(o + blk_x_idx, blk_grad, mask=blk_x_mask)


def softmax_fused_build_lut(lut: dict, sparsity_layout: Tensor):
    if lut is None:
        lut = dict()

    if "sparsity_reverse_lut_sorted" not in lut:
        sparsity_layout_flat = sparsity_layout.reshape(-1)
        sparsity_reverse_lut_sorted = (((torch.cumsum(sparsity_layout_flat, dim=-1) - 1) *
                                        (sparsity_layout_flat == 1) -
                                        (1 * (sparsity_layout_flat == 0)))
                                       .reshape(sparsity_layout.size())
                                       .sort(descending=True, dim=-1)[0]
                                       .reshape(-1).contiguous())
        lut["sparsity_reverse_lut_sorted"] = sparsity_reverse_lut_sorted

    if "max_blocks_line" not in lut:
        sparsity_reverse_lut_sorted = lut["sparsity_reverse_lut_sorted"]
        max_blocks_line = ((torch.reshape(sparsity_reverse_lut_sorted, (-1, sparsity_layout.size(-1)))
                            != -1)
                           .sum(dim=-1)
                           .max()
                           .item())
        lut["max_blocks_line"] = ceil_pow2(max(max_blocks_line, 2))

    validate_contiguous(sparsity_layout, lut["sparsity_reverse_lut_sorted"])

    return lut


# noinspection PyUnusedLocal
def softmax_fused_setup_context(ctx, inputs, output):
    (_, sparsity_layout, sparsity_reverse_lut_sorted, max_blocks_line, sparsity_block_size) = inputs

    ctx.save_for_backward(output, sparsity_layout, sparsity_reverse_lut_sorted)
    ctx.max_blocks_line = max_blocks_line
    ctx.sparsity_block_size = sparsity_block_size


softmax_fused_forward.register_autograd(softmax_fused_backward_wrapper, setup_context=softmax_fused_setup_context)



"""Block-sparse Flash Attention implementation for blksprs.

This module implements Flash Attention 2 algorithm with block-sparse support,
including cross-attention (seq_q != seq_k) and custom attention masks.

Note: This implementation was developed with AI assistance.
"""

import math
from typing import Tuple

import torch
import triton
from torch import Tensor
from triton import language as tl

from blksprs.utils.validation import validate_contiguous, validate_device, validate_dtype_float, ensure_contiguous


@torch.amp.custom_fwd(device_type="cuda", cast_inputs=torch.float16)
def flash_attention(
    q: Tensor,
    k: Tensor,
    v: Tensor,
    attention_layout: Tensor,
    sparsity_block_size: int,
    scale: float = None,
    attention_mask: Tensor = None,
    lut: dict = None,
) -> Tensor:
    """Block-sparse flash attention with optional attention mask.
    
    Args:
        q: Query tensor [batch, seq_q, n_heads, head_dim]
        k: Key tensor [batch, seq_k, n_heads, head_dim]
        v: Value tensor [batch, seq_k, n_heads, head_dim]
        attention_layout: Block attention pattern [batch*heads, n_seq_blocks_q, n_seq_blocks_k]
        sparsity_block_size: Block size for sparsity pattern
        scale: Attention scale (default: 1/sqrt(head_dim))
        attention_mask: Boolean mask [batch*heads, seq_q, seq_k] where True=masked (default None)
        lut: Optional pre-computed LUT dictionary
        
    Returns:
        Output tensor [batch, seq_q, n_heads, head_dim]
    """
    q, k, v = ensure_contiguous(q, k, v)
    
    validate_contiguous(q, k, v)
    validate_dtype_float(q, k, v)
    validate_device(q, k, v)
    
    batch, seq_q, n_heads, head_dim = q.shape
    _, seq_k, _, _ = k.shape
    
    if k.shape[0] != batch or k.shape[2] != n_heads or k.shape[3] != head_dim:
        raise ValueError("K must have compatible shape with Q")
    if v.shape != k.shape:
        raise ValueError("V must have same shape as K")
    if not (sparsity_block_size >= 16 and (sparsity_block_size & (sparsity_block_size - 1)) == 0):
        raise ValueError(f"sparsity_block_size must be power of 2 >= 16, got {sparsity_block_size}")
    if seq_q % sparsity_block_size != 0:
        raise ValueError(f"seq_q ({seq_q}) must be divisible by sparsity_block_size")
    if seq_k % sparsity_block_size != 0:
        raise ValueError(f"seq_k ({seq_k}) must be divisible by sparsity_block_size")
    
    n_batches = batch * n_heads
    n_seq_blocks_q = seq_q // sparsity_block_size
    n_seq_blocks_k = seq_k // sparsity_block_size
    
    expected_layout_shape = (n_batches, n_seq_blocks_q, n_seq_blocks_k)
    if attention_layout.shape != expected_layout_shape:
        raise ValueError(f"attention_layout shape {tuple(attention_layout.shape)} doesn't match expected {expected_layout_shape}")
    
    if scale is None:
        scale = 1.0 / math.sqrt(head_dim)
    
    if lut is None:
        lut = flash_attention_build_lut(attention_layout, n_seq_blocks_q, n_seq_blocks_k)
    
    has_mask = attention_mask is not None
    if has_mask:
        if attention_mask.shape != (n_batches, seq_q, seq_k):
            raise ValueError(f"attention_mask shape {tuple(attention_mask.shape)} doesn't match expected ({n_batches}, {seq_q}, {seq_k})")
        attention_mask_additive = torch.where(
            attention_mask, 
            torch.tensor(float("-inf"), device=attention_mask.device, dtype=q.dtype),
            torch.tensor(0.0, device=attention_mask.device, dtype=q.dtype)
        ).contiguous()
    else:
        attention_mask_additive = torch.empty(0, device=q.device, dtype=q.dtype)
    
    return BlockSparseFlashAttention.apply(
        q, k, v,
        attention_mask_additive,
        lut["attn_lut"], lut["attn_offsets"],
        lut["rev_attn_lut"], lut["rev_attn_offsets"],
        sparsity_block_size, n_seq_blocks_q, n_seq_blocks_k,
        lut["max_kv_blocks"], lut["max_q_per_k"],
        scale, has_mask,
    )


class BlockSparseFlashAttention(torch.autograd.Function):
    """Block-sparse Flash Attention with autograd support."""
    
    @staticmethod
    def forward(ctx, q, k, v, attention_mask, attn_lut, attn_offsets, rev_attn_lut, rev_attn_offsets,
                sparsity_block_size, n_seq_blocks_q, n_seq_blocks_k, max_kv_blocks, max_q_per_k, scale, has_mask):
        batch, seq_q, n_heads, head_dim = q.shape
        _, seq_k, _, _ = k.shape
        n_batches = batch * n_heads
        
        q_flat = q.permute(0, 2, 1, 3).reshape(n_batches, seq_q, head_dim).contiguous()
        k_flat = k.permute(0, 2, 1, 3).reshape(n_batches, seq_k, head_dim).contiguous()
        v_flat = v.permute(0, 2, 1, 3).reshape(n_batches, seq_k, head_dim).contiguous()
        
        o_flat = torch.empty_like(q_flat)
        lse = torch.empty(n_batches, seq_q, device=q.device, dtype=torch.float32)
        l = torch.empty(n_batches, seq_q, device=q.device, dtype=torch.float32)
        
        if head_dim <= 64:
            BLOCK_M = min(128, sparsity_block_size)
        elif head_dim <= 128:
            BLOCK_M = min(64, sparsity_block_size)
        else:
            BLOCK_M = min(32, sparsity_block_size)
        BLOCK_N = sparsity_block_size
        
        n_m_tiles = seq_q // BLOCK_M
        grid = (n_m_tiles, n_batches)
        
        if has_mask:
            mask_stride_batch = attention_mask.stride(0)
            mask_stride_row = attention_mask.stride(1)
            mask_stride_col = attention_mask.stride(2)
        else:
            mask_stride_batch = 0
            mask_stride_row = 0
            mask_stride_col = 0
        
        flash_attention_fwd_kernel[grid](
            q_flat, k_flat, v_flat, o_flat,
            attention_mask if has_mask else q_flat,
            attn_lut, attn_offsets,
            lse, l,
            q_flat.stride(0), q_flat.stride(1), q_flat.stride(2),
            k_flat.stride(0), k_flat.stride(1), k_flat.stride(2),
            mask_stride_batch, mask_stride_row, mask_stride_col,
            n_batches, seq_q, seq_k, head_dim, sparsity_block_size, n_seq_blocks_q, max_kv_blocks,
            scale,
            has_mask,
            BLOCK_M=BLOCK_M, BLOCK_N=BLOCK_N,
            num_stages=4, num_warps=4,
        )
        
        o = o_flat.reshape(batch, n_heads, seq_q, head_dim).permute(0, 2, 1, 3).contiguous()
        
        ctx.save_for_backward(q_flat, k_flat, v_flat, o_flat, lse,
                               attn_lut, attn_offsets, rev_attn_lut, rev_attn_offsets,
                               attention_mask if has_mask else torch.empty(0, device=q.device))
        ctx.sparsity_block_size = sparsity_block_size
        ctx.n_seq_blocks_q = n_seq_blocks_q
        ctx.n_seq_blocks_k = n_seq_blocks_k
        ctx.max_kv_blocks = max_kv_blocks
        ctx.max_q_per_k = max_q_per_k
        ctx.scale = scale
        ctx.has_mask = has_mask
        ctx.batch = batch
        ctx.n_heads = n_heads
        ctx.seq_q = seq_q
        ctx.seq_k = seq_k
        ctx.head_dim = head_dim
        ctx.BLOCK_M = BLOCK_M
        ctx.BLOCK_N = BLOCK_N
        
        return o
    
    @staticmethod
    def backward(ctx, grad_output):
        (q_flat, k_flat, v_flat, o_flat, lse,
         attn_lut, attn_offsets, rev_attn_lut, rev_attn_offsets, attention_mask) = ctx.saved_tensors
        
        batch = ctx.batch
        n_heads = ctx.n_heads
        seq_q = ctx.seq_q
        seq_k = ctx.seq_k
        head_dim = ctx.head_dim
        n_batches = batch * n_heads
        sparsity_block_size = ctx.sparsity_block_size
        BLOCK_M = ctx.BLOCK_M
        BLOCK_N = ctx.BLOCK_N
        has_mask = ctx.has_mask
        
        do_flat = grad_output.permute(0, 2, 1, 3).reshape(n_batches, seq_q, head_dim).contiguous()
        
        dq_flat = torch.zeros_like(q_flat)
        dk_flat = torch.zeros_like(k_flat)
        dv_flat = torch.zeros_like(v_flat)
        delta = torch.empty(n_batches, seq_q, device=q_flat.device, dtype=torch.float32)
        
        if has_mask:
            mask_stride_batch = attention_mask.stride(0)
            mask_stride_row = attention_mask.stride(1)
            mask_stride_col = attention_mask.stride(2)
        else:
            mask_stride_batch = 0
            mask_stride_row = 0
            mask_stride_col = 0
        
        n_m_tiles_q = seq_q // BLOCK_M
        flash_attention_bwd_preprocess_kernel[(n_m_tiles_q, n_batches)](
            o_flat, do_flat, delta,
            o_flat.stride(0), o_flat.stride(1), o_flat.stride(2),
            seq_q, head_dim,
            BLOCK_M=BLOCK_M,
        )
        
        n_n_tiles_k = seq_k // BLOCK_N
        flash_attention_bwd_dkdv_kernel[(n_n_tiles_k, n_batches)](
            q_flat, k_flat, v_flat, do_flat,
            dk_flat, dv_flat,
            lse, delta,
            attention_mask if has_mask else q_flat,
            rev_attn_lut, rev_attn_offsets,
            q_flat.stride(0), q_flat.stride(1),
            k_flat.stride(0), k_flat.stride(1),
            q_flat.stride(2),
            mask_stride_batch, mask_stride_row, mask_stride_col,
            n_batches, seq_q, seq_k, head_dim, sparsity_block_size, ctx.n_seq_blocks_k, ctx.max_q_per_k,
            ctx.scale,
            has_mask,
            BLOCK_M=BLOCK_M, BLOCK_N=BLOCK_N,
        )
        
        flash_attention_bwd_dq_kernel[(n_m_tiles_q, n_batches)](
            q_flat, k_flat, v_flat, do_flat,
            dq_flat,
            lse, delta,
            attention_mask if has_mask else q_flat,
            attn_lut, attn_offsets,
            q_flat.stride(0), q_flat.stride(1),
            k_flat.stride(0), k_flat.stride(1),
            q_flat.stride(2),
            mask_stride_batch, mask_stride_row, mask_stride_col,
            n_batches, seq_q, seq_k, head_dim, sparsity_block_size, ctx.n_seq_blocks_q, ctx.max_kv_blocks,
            ctx.scale,
            has_mask,
            BLOCK_M=BLOCK_M, BLOCK_N=BLOCK_N,
        )
        
        dq = dq_flat.reshape(batch, n_heads, seq_q, head_dim).permute(0, 2, 1, 3).contiguous()
        dk = dk_flat.reshape(batch, n_heads, seq_k, head_dim).permute(0, 2, 1, 3).contiguous()
        dv = dv_flat.reshape(batch, n_heads, seq_k, head_dim).permute(0, 2, 1, 3).contiguous()
        
        return dq, dk, dv, None, None, None, None, None, None, None, None, None, None, None, None


@triton.jit
def flash_attention_fwd_kernel(
    q_ptr, k_ptr, v_ptr, o_ptr,
    mask_ptr,
    attn_lut_ptr, attn_offsets_ptr,
    m_ptr, l_ptr,
    stride_q_batch, stride_q_seq, stride_q_dim,
    stride_kv_batch, stride_kv_seq, stride_kv_dim,
    stride_mask_batch, stride_mask_row, stride_mask_col,
    n_batches: tl.constexpr,
    seq_q: tl.constexpr,
    seq_k: tl.constexpr,
    head_dim: tl.constexpr,
    sparsity_block_size: tl.constexpr,
    n_seq_blocks_q: tl.constexpr,
    max_kv_blocks: tl.constexpr,
    scale,
    has_mask: tl.constexpr,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    """Flash attention forward kernel with block-sparse mask support."""
    pid_m = tl.program_id(0)
    pid_batch = tl.program_id(1)
    
    n_m_tiles: tl.constexpr = sparsity_block_size // BLOCK_M
    n_n_tiles: tl.constexpr = sparsity_block_size // BLOCK_N
    
    q_seq_block = pid_m // n_m_tiles
    m_tile_idx = pid_m % n_m_tiles
    
    q_row_start = q_seq_block * sparsity_block_size + m_tile_idx * BLOCK_M
    offs_m = q_row_start + tl.arange(0, BLOCK_M)
    offs_d = tl.arange(0, head_dim)
    
    q_ptrs = q_ptr + pid_batch * stride_q_batch + offs_m[:, None] * stride_q_seq + offs_d[None, :]
    q_mask = offs_m[:, None] < seq_q
    q = tl.load(q_ptrs, mask=q_mask, other=0.0)
    
    m_i = tl.full([BLOCK_M], float("-inf"), dtype=tl.float32)
    l_i = tl.zeros([BLOCK_M], dtype=tl.float32)
    acc = tl.zeros([BLOCK_M, head_dim], dtype=tl.float32)
    
    qk_scale = scale * 1.44269504
    
    attn_offset_idx = pid_batch * n_seq_blocks_q + q_seq_block
    attn_start = tl.load(attn_offsets_ptr + attn_offset_idx)
    attn_end = tl.load(attn_offsets_ptr + attn_offset_idx + 1)
    n_kv_blocks = attn_end - attn_start
    
    for kv_idx in range(max_kv_blocks):
        if kv_idx < n_kv_blocks:
            k_seq_block = tl.load(attn_lut_ptr + attn_start + kv_idx)
            
            k_row_start = k_seq_block * sparsity_block_size
            offs_n = k_row_start + tl.arange(0, BLOCK_N)
            
            k_ptrs = k_ptr + pid_batch * stride_kv_batch + offs_n[:, None] * stride_kv_seq + offs_d[None, :]
            k_mask = offs_n[:, None] < seq_k
            k = tl.load(k_ptrs, mask=k_mask, other=0.0)
            
            qk = tl.dot(q, tl.trans(k)) * qk_scale
            
            if has_mask:
                mask_ptrs = mask_ptr + pid_batch * stride_mask_batch + offs_m[:, None] * stride_mask_row + offs_n[None, :] * stride_mask_col
                mask_vals = tl.load(mask_ptrs, mask=(offs_m[:, None] < seq_q) & (offs_n[None, :] < seq_k), other=0.0)
                qk = qk + mask_vals
            
            m_ij = tl.maximum(m_i, tl.max(qk, axis=1))
            alpha = tl.math.exp2(m_i - m_ij)
            p = tl.math.exp2(qk - m_ij[:, None])
            l_i = l_i * alpha + tl.sum(p, axis=1)
            acc = acc * alpha[:, None]
            
            v_ptrs = v_ptr + pid_batch * stride_kv_batch + offs_n[:, None] * stride_kv_seq + offs_d[None, :]
            v = tl.load(v_ptrs, mask=k_mask, other=0.0)
            acc = tl.dot(p.to(v.dtype), v, acc)
            
            m_i = m_ij
    
    has_attention = l_i > 0
    l_safe = tl.where(has_attention, l_i, 1.0)
    acc = acc / l_safe[:, None]
    acc = tl.where(has_attention[:, None], acc, 0.0)
    
    o_ptrs = o_ptr + pid_batch * stride_q_batch + offs_m[:, None] * stride_q_seq + offs_d[None, :]
    tl.store(o_ptrs, acc.to(o_ptr.dtype.element_ty), mask=offs_m[:, None] < seq_q)
    
    lse = tl.where(has_attention, m_i + tl.math.log2(l_safe), float("-inf"))
    tl.store(m_ptr + pid_batch * seq_q + offs_m, lse, mask=offs_m < seq_q)
    tl.store(l_ptr + pid_batch * seq_q + offs_m, l_i, mask=offs_m < seq_q)


@triton.jit
def flash_attention_bwd_preprocess_kernel(
    o_ptr, do_ptr, delta_ptr,
    stride_batch, stride_seq, stride_dim,
    seq_len: tl.constexpr,
    head_dim: tl.constexpr,
    BLOCK_M: tl.constexpr,
):
    """Compute delta = (O * dO).sum(dim=-1)."""
    pid_m = tl.program_id(0)
    pid_batch = tl.program_id(1)
    
    offs_m = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    offs_d = tl.arange(0, head_dim)
    
    o_ptrs = o_ptr + pid_batch * stride_batch + offs_m[:, None] * stride_seq + offs_d[None, :]
    do_ptrs = do_ptr + pid_batch * stride_batch + offs_m[:, None] * stride_seq + offs_d[None, :]
    mask = offs_m[:, None] < seq_len
    
    o = tl.load(o_ptrs, mask=mask, other=0.0).to(tl.float32)
    do = tl.load(do_ptrs, mask=mask, other=0.0).to(tl.float32)
    delta = tl.sum(o * do, axis=1)
    
    tl.store(delta_ptr + pid_batch * seq_len + offs_m, delta, mask=offs_m < seq_len)


@triton.jit
def flash_attention_bwd_dkdv_kernel(
    q_ptr, k_ptr, v_ptr, do_ptr,
    dk_ptr, dv_ptr,
    lse_ptr, delta_ptr,
    mask_ptr,
    rev_attn_lut_ptr, rev_attn_offsets_ptr,
    stride_q_batch, stride_q_seq,
    stride_kv_batch, stride_kv_seq,
    stride_dim,
    stride_mask_batch, stride_mask_row, stride_mask_col,
    n_batches: tl.constexpr,
    seq_q: tl.constexpr,
    seq_k: tl.constexpr,
    head_dim: tl.constexpr,
    sparsity_block_size: tl.constexpr,
    n_seq_blocks_k: tl.constexpr,
    max_q_per_k: tl.constexpr,
    scale,
    has_mask: tl.constexpr,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    """Compute dK and dV gradients."""
    pid_n = tl.program_id(0)
    pid_batch = tl.program_id(1)
    
    n_n_tiles = sparsity_block_size // BLOCK_N
    n_m_tiles = sparsity_block_size // BLOCK_M
    
    k_seq_block = pid_n // n_n_tiles
    n_tile_idx = pid_n % n_n_tiles
    
    k_row_start = k_seq_block * sparsity_block_size + n_tile_idx * BLOCK_N
    offs_n = k_row_start + tl.arange(0, BLOCK_N)
    offs_d = tl.arange(0, head_dim)
    
    qk_scale = scale * 1.44269504
    
    k_ptrs = k_ptr + pid_batch * stride_kv_batch + offs_n[:, None] * stride_kv_seq + offs_d[None, :]
    v_ptrs = v_ptr + pid_batch * stride_kv_batch + offs_n[:, None] * stride_kv_seq + offs_d[None, :]
    k_mask = offs_n[:, None] < seq_k
    k = tl.load(k_ptrs, mask=k_mask, other=0.0)
    v = tl.load(v_ptrs, mask=k_mask, other=0.0)
    
    dk = tl.zeros([BLOCK_N, head_dim], dtype=tl.float32)
    dv = tl.zeros([BLOCK_N, head_dim], dtype=tl.float32)
    
    rev_offset_idx = pid_batch * n_seq_blocks_k + k_seq_block
    rev_start = tl.load(rev_attn_offsets_ptr + rev_offset_idx)
    rev_end = tl.load(rev_attn_offsets_ptr + rev_offset_idx + 1)
    n_q_blocks = rev_end - rev_start
    
    for q_idx in range(max_q_per_k):
        if q_idx < n_q_blocks:
            q_seq_block = tl.load(rev_attn_lut_ptr + rev_start + q_idx)
            
            for m_tile_idx in range(n_m_tiles):
                q_row_start = q_seq_block * sparsity_block_size + m_tile_idx * BLOCK_M
                offs_m = q_row_start + tl.arange(0, BLOCK_M)
                
                q_ptrs = q_ptr + pid_batch * stride_q_batch + offs_m[:, None] * stride_q_seq + offs_d[None, :]
                do_ptrs = do_ptr + pid_batch * stride_q_batch + offs_m[:, None] * stride_q_seq + offs_d[None, :]
                q_mask = offs_m[:, None] < seq_q
                q = tl.load(q_ptrs, mask=q_mask, other=0.0)
                do = tl.load(do_ptrs, mask=q_mask, other=0.0)
                
                m = tl.load(lse_ptr + pid_batch * seq_q + offs_m, mask=offs_m < seq_q, other=0.0)
                Di = tl.load(delta_ptr + pid_batch * seq_q + offs_m, mask=offs_m < seq_q, other=0.0)
                
                qk = tl.dot(q, tl.trans(k)) * qk_scale
                
                if has_mask:
                    mask_ptrs = mask_ptr + pid_batch * stride_mask_batch + offs_m[:, None] * stride_mask_row + offs_n[None, :] * stride_mask_col
                    mask_vals = tl.load(mask_ptrs, mask=(offs_m[:, None] < seq_q) & (offs_n[None, :] < seq_k), other=0.0)
                    qk = qk + mask_vals
                
                valid_lse = m > float("-inf")
                safe_m = tl.where(valid_lse, m, 0.0)
                p = tl.math.exp2(qk - safe_m[:, None])
                p = tl.where(valid_lse[:, None], p, 0.0)
                
                dv += tl.dot(tl.trans(p.to(do.dtype)), do)
                dp = tl.dot(do, tl.trans(v))
                ds = p * (dp - Di[:, None])
                dk += tl.dot(tl.trans(ds.to(q.dtype)), q)
    
    dk = dk * scale
    tl.store(dk_ptr + pid_batch * stride_kv_batch + offs_n[:, None] * stride_kv_seq + offs_d[None, :], dk.to(dk_ptr.dtype.element_ty), mask=k_mask)
    tl.store(dv_ptr + pid_batch * stride_kv_batch + offs_n[:, None] * stride_kv_seq + offs_d[None, :], dv.to(dv_ptr.dtype.element_ty), mask=k_mask)


@triton.jit
def flash_attention_bwd_dq_kernel(
    q_ptr, k_ptr, v_ptr, do_ptr,
    dq_ptr,
    lse_ptr, delta_ptr,
    mask_ptr,
    attn_lut_ptr, attn_offsets_ptr,
    stride_q_batch, stride_q_seq,
    stride_kv_batch, stride_kv_seq,
    stride_dim,
    stride_mask_batch, stride_mask_row, stride_mask_col,
    n_batches: tl.constexpr,
    seq_q: tl.constexpr,
    seq_k: tl.constexpr,
    head_dim: tl.constexpr,
    sparsity_block_size: tl.constexpr,
    n_seq_blocks_q: tl.constexpr,
    max_kv_blocks: tl.constexpr,
    scale,
    has_mask: tl.constexpr,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    """Compute dQ gradients."""
    pid_m = tl.program_id(0)
    pid_batch = tl.program_id(1)
    
    n_m_tiles = sparsity_block_size // BLOCK_M
    n_n_tiles = sparsity_block_size // BLOCK_N
    
    q_seq_block = pid_m // n_m_tiles
    m_tile_idx = pid_m % n_m_tiles
    
    q_row_start = q_seq_block * sparsity_block_size + m_tile_idx * BLOCK_M
    offs_m = q_row_start + tl.arange(0, BLOCK_M)
    offs_d = tl.arange(0, head_dim)
    
    qk_scale = scale * 1.44269504
    
    q_ptrs = q_ptr + pid_batch * stride_q_batch + offs_m[:, None] * stride_q_seq + offs_d[None, :]
    do_ptrs = do_ptr + pid_batch * stride_q_batch + offs_m[:, None] * stride_q_seq + offs_d[None, :]
    q_mask = offs_m[:, None] < seq_q
    q = tl.load(q_ptrs, mask=q_mask, other=0.0)
    do = tl.load(do_ptrs, mask=q_mask, other=0.0)
    
    m = tl.load(lse_ptr + pid_batch * seq_q + offs_m, mask=offs_m < seq_q, other=0.0)
    Di = tl.load(delta_ptr + pid_batch * seq_q + offs_m, mask=offs_m < seq_q, other=0.0)
    
    dq = tl.zeros([BLOCK_M, head_dim], dtype=tl.float32)
    
    attn_offset_idx = pid_batch * n_seq_blocks_q + q_seq_block
    attn_start = tl.load(attn_offsets_ptr + attn_offset_idx)
    attn_end = tl.load(attn_offsets_ptr + attn_offset_idx + 1)
    n_kv_blocks = attn_end - attn_start
    
    for kv_idx in range(max_kv_blocks):
        if kv_idx < n_kv_blocks:
            k_seq_block = tl.load(attn_lut_ptr + attn_start + kv_idx)
            
            for n_tile_idx in range(n_n_tiles):
                k_row_start = k_seq_block * sparsity_block_size + n_tile_idx * BLOCK_N
                offs_n = k_row_start + tl.arange(0, BLOCK_N)
                
                k_ptrs = k_ptr + pid_batch * stride_kv_batch + offs_n[:, None] * stride_kv_seq + offs_d[None, :]
                v_ptrs = v_ptr + pid_batch * stride_kv_batch + offs_n[:, None] * stride_kv_seq + offs_d[None, :]
                k_mask = offs_n[:, None] < seq_k
                k = tl.load(k_ptrs, mask=k_mask, other=0.0)
                v = tl.load(v_ptrs, mask=k_mask, other=0.0)
                
                qk = tl.dot(q, tl.trans(k)) * qk_scale
                
                if has_mask:
                    mask_ptrs = mask_ptr + pid_batch * stride_mask_batch + offs_m[:, None] * stride_mask_row + offs_n[None, :] * stride_mask_col
                    mask_vals = tl.load(mask_ptrs, mask=(offs_m[:, None] < seq_q) & (offs_n[None, :] < seq_k), other=0.0)
                    qk = qk + mask_vals
                
                valid_lse = m > float("-inf")
                safe_m = tl.where(valid_lse, m, 0.0)
                p = tl.math.exp2(qk - safe_m[:, None])
                p = tl.where(valid_lse[:, None], p, 0.0)
                
                dp = tl.dot(do, tl.trans(v))
                ds = p * (dp - Di[:, None])
                dq += tl.dot(ds.to(k.dtype), k)
    
    dq = dq * scale
    tl.store(dq_ptr + pid_batch * stride_q_batch + offs_m[:, None] * stride_q_seq + offs_d[None, :], dq.to(dq_ptr.dtype.element_ty), mask=q_mask)


def flash_attention_build_lut(
    attention_layout: Tensor,
    n_seq_blocks_q: int = None,
    n_seq_blocks_k: int = None,
) -> dict:
    """Build attention LUTs for reuse across multiple calls."""
    n_batches = attention_layout.shape[0]
    if n_seq_blocks_q is None:
        n_seq_blocks_q = attention_layout.shape[1]
    if n_seq_blocks_k is None:
        n_seq_blocks_k = attention_layout.shape[2]
    
    attn_lut, attn_offsets, max_kv_blocks = _build_attention_lut_fast(
        attention_layout, n_batches, n_seq_blocks_q, n_seq_blocks_k
    )
    
    attention_layout_t = attention_layout.transpose(1, 2).contiguous()
    rev_attn_lut, rev_attn_offsets, max_q_per_k = _build_attention_lut_fast(
        attention_layout_t, n_batches, n_seq_blocks_k, n_seq_blocks_q
    )
    
    return {
        "attn_lut": attn_lut,
        "attn_offsets": attn_offsets,
        "max_kv_blocks": max_kv_blocks,
        "rev_attn_lut": rev_attn_lut,
        "rev_attn_offsets": rev_attn_offsets,
        "max_q_per_k": max_q_per_k,
    }


def _build_attention_lut_fast(
    attention_layout: Tensor,
    n_batches: int,
    n_blocks_row: int,
    n_blocks_col: int,
) -> Tuple[Tensor, Tensor, int]:
    """Build attention LUT efficiently."""
    device = attention_layout.device
    
    counts = attention_layout.sum(dim=2).flatten()
    max_blocks_per_row = int(counts.max().item())
    
    if max_blocks_per_row == 0:
        offsets = torch.zeros(n_batches * n_blocks_row + 1, dtype=torch.int32, device=device)
        lut = torch.empty(0, dtype=torch.int32, device=device)
        return lut, offsets, 1
    
    offsets = torch.zeros(n_batches * n_blocks_row + 1, dtype=torch.int32, device=device)
    offsets[1:] = counts.cumsum(0).to(torch.int32)
    
    indices = attention_layout.reshape(n_batches * n_blocks_row, n_blocks_col).nonzero(as_tuple=False)
    lut = indices[:, 1].to(torch.int32)
    
    return lut, offsets, max_blocks_per_row



import torch
import triton
from torch import Tensor
from torch._library.triton import wrap_triton, triton_op
from triton import language as tl

from blksprs.utils.autotuning import get_autotune_configs, prune_autotune_configs
from blksprs.utils.blksprs_tensor import BlksprsTensor
from blksprs.utils.tools import stride
from blksprs.utils.validation import validate_dimensions, validate_contiguous, validate_device, validate_sparsity, \
    validate_sparsity_block_size, ensure_contiguous


@torch.amp.custom_fwd(device_type="cuda", cast_inputs=torch.float32)
def row_wise_sum(x: BlksprsTensor, sparsity_layout: Tensor, sparsity_block_size: int,
                 flag_slice_only: bool = False) -> (BlksprsTensor, Tensor):
    """Computes the row-wise sum of a block-sparse tensor.

    Returns a block-sparse tensor in compressed form with only one block per row, where the first entry contains the sum
        of the corresponding row.

    Note:
        If ``flag_slice_only`` is set the output will be of shape ``[x.size(0), x.size(1), 1]``.

    Args:
        x (BlksprsTensor): A block-sparse tensor in compressed form.
        sparsity_layout (Tensor): The sparsity layout of the block-sparse tensor.
        sparsity_block_size (int): The size of the sparsity blocks.
        flag_slice_only (bool, optional): If set the output will be of shape ``[x.size(0), x.size(1), 1]``
            (default ``False``).

    Returns:
        tuple[BlksprsTensor, Tensor]: A tuple containing a block-sparse tensor in compressed form containing the row-wise sum
            of the input and the sparsity layout of the output tensor.

    """
    x = ensure_contiguous(x)

    validate_dimensions(x)
    validate_contiguous(x)
    validate_device(x)
    validate_sparsity(sparsity_block_size, (x, sparsity_layout))
    validate_sparsity_block_size(sparsity_block_size, x)

    sparsity_lut = torch.nonzero(sparsity_layout).contiguous()

    sparsity_layout_output, _ = torch.max(sparsity_layout, dim=-1, keepdim=True)
    sparsity_layout_output_flat = sparsity_layout_output.reshape(-1)
    sparsity_reverse_lut_output = ((torch.cumsum(sparsity_layout_output_flat, dim=-1) - 1) *
                                   (sparsity_layout_output_flat == 1) -
                                   (1 * (sparsity_layout_output_flat == 0)))

    n_sparse_blocks_output = torch.sum(sparsity_layout_output.to(torch.int)).item()

    validate_contiguous(sparsity_layout, sparsity_lut,
                        sparsity_layout_output, sparsity_reverse_lut_output)

    return BlksprsTensor.wrap(row_wise_sum_forward(
        x, sparsity_lut, sparsity_layout_output, sparsity_reverse_lut_output,
        sparsity_block_size, n_sparse_blocks_output, flag_slice_only)), sparsity_layout_output


@triton_op("blksprs::row_wise_sum_forward", mutates_args={})
def row_wise_sum_forward(x: Tensor, sparsity_lut: Tensor,
                         sparsity_layout_output: Tensor, sparsity_reverse_lut_output: Tensor,
                         sparsity_block_size: int, n_sparse_blocks_output: int,
                         flag_slice_only: bool = False) -> Tensor:
    with torch.no_grad():
        output = torch.zeros(
            size=(n_sparse_blocks_output, sparsity_block_size, 1 if flag_slice_only else sparsity_block_size),
            dtype=x.dtype, device=x.device)

        x_b, x_r, x_c = x.size()
        x_b_s, x_r_s, x_c_s = stride(x)
        s_lut_x_r, s_lut_x_c = sparsity_lut.size()
        s_lut_x_r_s, s_lut_x_c_s = stride(sparsity_lut)
        o_b, o_r, o_c = output.size()
        o_b_s, o_r_s, o_c_s = stride(output)
        s_l_o_b, s_l_o_r, s_l_o_c = sparsity_layout_output.size()
        s_l_o_b_s, s_l_o_r_s, s_l_o_c_s = stride(sparsity_layout_output)

        triton_grid = lambda meta: [x_b,
                                    triton.cdiv(x_r, meta["TRITON_BLOCK_SIZE"]),
                                    triton.cdiv(x_c, meta["TRITON_BLOCK_SIZE"])]

        (wrap_triton(row_wise_sum_kernel)[triton_grid]
         (x,
          x_b, x_b_s, x_r_s, x_c_s,
          sparsity_lut, s_lut_x_r, s_lut_x_r_s, s_lut_x_c_s,
          output,
          o_b, o_b_s, o_r_s,
          s_l_o_b, s_l_o_b_s, s_l_o_r_s,
          sparsity_reverse_lut_output,
          sparsity_block_size))

        return output


# noinspection PyUnusedLocal
@triton.autotune(
    configs=get_autotune_configs(),
    key=["sparsity_block_size"],
    prune_configs_by={"early_config_prune": prune_autotune_configs},
    reset_to_zero=["o"]
)
@triton.jit
def row_wise_sum_kernel(x,
                        x_b, x_b_s, x_r_s, x_c_s,
                        s_lut_x, s_lut_x_r, s_lut_x_r_s, s_lut_x_c_s,
                        o,
                        o_b, o_b_s, o_r_s,
                        s_l_o_b, s_l_o_b_s, s_l_o_r_s,
                        r_lut_o,
                        sparsity_block_size,
                        TRITON_BLOCK_SIZE: tl.constexpr) -> None:
    # Get triton block indices
    pid_blk = tl.program_id(axis=0)
    pid_row = tl.program_id(axis=1)
    pid_col = tl.program_id(axis=2)

    # Get position of current sparsity block consisting of its batch and row index
    spa_val_idx = pid_blk * s_lut_x_r_s + tl.arange(0, 4) * s_lut_x_c_s
    spa_val_msk = (tl.arange(0, 4) < 3)
    spa_val = tl.load(s_lut_x + spa_val_idx, mask=spa_val_msk)

    spa_bat_x = tl.sum(spa_val * (tl.arange(0, 4) == 0))
    spa_row_x = tl.sum(spa_val * (tl.arange(0, 4) == 1))
    spa_col_x = tl.sum(spa_val * (tl.arange(0, 4) == 2))

    # Load reverse sparsity index for current block
    rev_idx_spa_idx = (spa_bat_x * s_l_o_b_s +
                       spa_row_x * s_l_o_r_s)
    rev_idx_spa_msk = ((rev_idx_spa_idx >= 0) &
                       (rev_idx_spa_idx < s_l_o_b * s_l_o_b_s))
    rev_idx_spa = tl.load(r_lut_o + rev_idx_spa_idx, mask=rev_idx_spa_msk).to(tl.int32)

    if rev_idx_spa >= 0:
        blk_idx = ((pid_blk * x_b_s) +
                   ((pid_row * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * x_r_s)[:, None] +
                   ((pid_col * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * x_c_s)[None, :])
        blk_msk = ((blk_idx >= 0) &
                   (blk_idx < x_b * x_b_s))
        blk = tl.load(x + blk_idx, mask=blk_msk)

        buf = tl.reshape(tl.sum(blk, axis=-1), (TRITON_BLOCK_SIZE, 1))

        o_idx = (rev_idx_spa * o_b_s +
                 ((pid_row * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * o_r_s)[:, None] +
                 (tl.arange(0, 1))[None, :])
        o_msk = ((o_idx >= 0) &
                 (o_idx < o_b * o_b_s))
        tl.atomic_add(o + o_idx, buf, o_msk)


@torch.amp.custom_fwd(device_type="cuda", cast_inputs=torch.float16)
def row_wise_max(x: BlksprsTensor, sparsity_layout: Tensor, sparsity_block_size: int,
                 flag_slice_only: bool = False) -> (BlksprsTensor, Tensor):
    """Computes the row-wise max of a block-sparse tensor.

    Returns a block-sparse tensor in compressed form with only one block per row, where the first entry contains the
        maximum of the corresponding row.

    Note:
        If ``flag_slice_only`` is set the output will be of shape ``[x.size(0), x.size(1), 1]``.

    Args:
        x (BlksprsTensor): A block-sparse tensor in compressed form.
        sparsity_layout (Tensor): The sparsity layout of the block-sparse tensor.
        sparsity_block_size (int): The size of the sparsity blocks.
        flag_slice_only (bool, optional): If set the output will be of shape ``[x.size(0), x.size(1), 1]``
            (default ``False``).

    Returns:
        tuple[BlksprsTensor, Tensor]: A tuple containing a block-sparse tensor in compressed form containing the row-wise max
            of the input and the sparsity layout of the output tensor.

    """
    x = ensure_contiguous(x)

    validate_dimensions(x)
    validate_contiguous(x)
    validate_device(x)
    validate_sparsity(sparsity_block_size, (x, sparsity_layout))
    validate_sparsity_block_size(sparsity_block_size, x)

    sparsity_lut = torch.nonzero(sparsity_layout).contiguous()

    sparsity_layout_output, _ = torch.max(sparsity_layout, dim=-1, keepdim=True)
    sparsity_layout_output_flat = sparsity_layout_output.reshape(-1)
    sparsity_reverse_lut_output = ((torch.cumsum(sparsity_layout_output_flat, dim=-1) - 1) *
                                   (sparsity_layout_output_flat == 1) -
                                   (1 * (sparsity_layout_output_flat == 0)))

    n_sparse_blocks_output = torch.sum(sparsity_layout_output.to(torch.int)).item()

    validate_contiguous(sparsity_layout, sparsity_lut,
                        sparsity_layout_output, sparsity_reverse_lut_output)

    return BlksprsTensor.wrap(
        row_wise_max_forward(x, sparsity_lut, sparsity_layout_output, sparsity_reverse_lut_output, sparsity_block_size,
                             n_sparse_blocks_output, flag_slice_only)), sparsity_layout_output


@triton_op("blksprs::row_wise_max_forward", mutates_args={})
def row_wise_max_forward(x: Tensor, sparsity_lut: Tensor,
                         sparsity_layout_output: Tensor, sparsity_reverse_lut_output: Tensor,
                         sparsity_block_size: int, n_sparse_blocks_output: int,
                         flag_slice_only: bool = False) -> Tensor:
    with torch.no_grad():
        output = torch.full(size=(n_sparse_blocks_output,
                                  sparsity_block_size,
                                  1 if flag_slice_only else sparsity_block_size),
                            fill_value=torch.finfo(x.dtype).min,
                            device=x.device)

        x_b, x_r, x_c = x.size()
        x_b_s, x_r_s, x_c_s = stride(x)
        s_lut_x_r, s_lut_x_c = sparsity_lut.size()
        s_lut_x_r_s, s_lut_x_c_s = stride(sparsity_lut)
        o_b, o_r, o_c = output.size()
        o_b_s, o_r_s, o_c_s = stride(output)
        s_l_o_b, s_l_o_r, s_l_o_c = sparsity_layout_output.size()
        s_l_o_b_s, s_l_o_r_s, s_l_o_c_s = stride(sparsity_layout_output)

        triton_grid = lambda meta: [x_b,
                                    triton.cdiv(x_r, meta["TRITON_BLOCK_SIZE"]),
                                    triton.cdiv(x_c, meta["TRITON_BLOCK_SIZE"])]

        (wrap_triton(row_wise_max_kernel)[triton_grid]
         (x,
          x_b, x_b_s, x_r_s, x_c_s,
          sparsity_lut, s_lut_x_r, s_lut_x_r_s, s_lut_x_c_s,
          output,
          o_b, o_b_s, o_r_s,
          s_l_o_b, s_l_o_b_s, s_l_o_r_s,
          sparsity_reverse_lut_output,
          sparsity_block_size))

        return output


# noinspection PyUnusedLocal
@triton.autotune(
    configs=get_autotune_configs(),
    key=["sparsity_block_size"],
    prune_configs_by={"early_config_prune": prune_autotune_configs},
    restore_value=["o"]
)
@triton.jit
def row_wise_max_kernel(x,
                        x_b, x_b_s, x_r_s, x_c_s,
                        s_lut_x, s_lut_x_r, s_lut_x_r_s, s_lut_x_c_s,
                        o,
                        o_b, o_b_s, o_r_s,
                        s_l_o_b, s_l_o_b_s, s_l_o_r_s,
                        r_lut_o,
                        sparsity_block_size,
                        TRITON_BLOCK_SIZE: tl.constexpr) -> None:
    # Get triton block indices
    pid_blk = tl.program_id(axis=0)
    pid_row = tl.program_id(axis=1)
    pid_col = tl.program_id(axis=2)

    # Get position of current sparsity block consisting of its batch and row index
    spa_val_idx = pid_blk * s_lut_x_r_s + tl.arange(0, 4) * s_lut_x_c_s
    spa_val_msk = (tl.arange(0, 4) < 3)
    spa_val = tl.load(s_lut_x + spa_val_idx, mask=spa_val_msk)

    spa_bat_x = tl.sum(spa_val * (tl.arange(0, 4) == 0))
    spa_row_x = tl.sum(spa_val * (tl.arange(0, 4) == 1))
    spa_col_x = tl.sum(spa_val * (tl.arange(0, 4) == 2))

    # Load reverse sparsity index for current block
    rev_idx_spa_idx = (spa_bat_x * s_l_o_b_s +
                       spa_row_x * s_l_o_r_s)
    rev_idx_spa_msk = ((rev_idx_spa_idx >= 0) &
                       (rev_idx_spa_idx < s_l_o_b * s_l_o_b_s))
    rev_idx_spa = tl.load(r_lut_o + rev_idx_spa_idx, mask=rev_idx_spa_msk).to(tl.int32)

    if rev_idx_spa >= 0:
        blk_idx = ((pid_blk * x_b_s) +
                   ((pid_row * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * x_r_s)[:, None] +
                   ((pid_col * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * x_c_s)[None, :])
        blk_msk = ((blk_idx >= 0) &
                   (blk_idx < x_b * x_b_s))
        blk = tl.load(x + blk_idx, mask=blk_msk)

        buf = tl.reshape(tl.max(blk, axis=-1), (TRITON_BLOCK_SIZE, 1))

        o_idx = (rev_idx_spa * o_b_s +
                 ((pid_row * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * o_r_s)[:, None] +
                 (tl.arange(0, 1))[None, :])
        o_msk = ((o_idx >= 0) &
                 (o_idx < o_b * o_b_s))
        tl.atomic_max(o + o_idx, buf, o_msk)


@torch.amp.custom_fwd(device_type="cuda", cast_inputs=torch.float16)
def row_wise_add(x: BlksprsTensor, sparsity_layout_x: Tensor, y: Tensor,
                 sparsity_block_size: int) -> BlksprsTensor:
    """For each row in ``y`` adds the value to each value in the corresponding row of the block-sparse tensor ``x``.

    Args:
        x (BlksprsTensor): A block-sparse tensor in compressed form.
        sparsity_layout_x (Tensor): The sparsity layout of the block-sparse tensor.
        y (BlksprsTensor): A block-sparse tensor in compressed form with only one value per row and a single column of sparse blocks.
        sparsity_block_size (int): The size of the sparsity blocks.

    Returns:
        BlksprsTensor: The values of ``x`` with the first value of ``y`` in each row added to them as a block-sparse tensor in
            compressed form.

    """
    x = ensure_contiguous(x)

    validate_dimensions(x)
    validate_contiguous(x)
    validate_device(x)
    validate_sparsity(sparsity_block_size, (x, sparsity_layout_x))
    validate_sparsity_block_size(sparsity_block_size, x)

    sparsity_lut_x = torch.nonzero(sparsity_layout_x).contiguous()

    sparsity_layout_rwm, _ = torch.max(sparsity_layout_x, dim=-1, keepdim=True)
    sparsity_layout_rwm_flat = sparsity_layout_rwm.reshape(-1)
    sparsity_reverse_lut_rwm = ((torch.cumsum(sparsity_layout_rwm_flat, dim=-1) - 1) *
                                (sparsity_layout_rwm_flat == 1) -
                                (1 * (sparsity_layout_rwm_flat == 0)))

    validate_contiguous(sparsity_layout_x, sparsity_lut_x, sparsity_reverse_lut_rwm)

    return BlksprsTensor.wrap(row_wise_add_forward(x, sparsity_lut_x, sparsity_layout_rwm,
                                                   sparsity_reverse_lut_rwm, y, sparsity_block_size))


def row_wise_sub(x: BlksprsTensor, sparsity_layout_x: Tensor, y: Tensor,
                 sparsity_block_size: int) -> BlksprsTensor:
    """Wrapper for ``row_wise_add`` with negated y.

    """
    return row_wise_add(x, sparsity_layout_x, torch.neg(y), sparsity_block_size)


@triton_op("blksprs::row_wise_add_forward", mutates_args={})
def row_wise_add_forward(x: Tensor, sparsity_lut_x: Tensor,
                         sparsity_layout_x_rwm: Tensor, sparsity_reverse_x_lut_rwm: Tensor,
                         y: Tensor, sparsity_block_size: int) -> Tensor:
    with torch.no_grad():
        output = torch.zeros_like(x)

        x_b, x_r, x_c = x.size()
        x_b_s, x_r_s, x_c_s = stride(x)
        s_lut_r, s_lut_c = sparsity_lut_x.size()
        s_lut_r_s, s_lut_c_s = stride(sparsity_lut_x)
        y_b, y_r, y_c = y.size()
        y_b_s, y_r_s, y_c_s = stride(y)
        s_l_y_b, s_l_y_r, s_l_y_c = sparsity_layout_x_rwm.size()
        s_l_y_b_s, s_l_y_r_s, s_l_y_c_s = stride(sparsity_layout_x_rwm)
        o_b, o_r, o_c = output.size()
        o_b_s, o_r_s, o_c_s = stride(output)

        triton_grid = lambda meta: [o_b,
                                    triton.cdiv(o_r, meta["TRITON_BLOCK_SIZE"]),
                                    triton.cdiv(o_c, meta["TRITON_BLOCK_SIZE"])]

        (wrap_triton(row_wise_add_kernel)[triton_grid]
         (x,
          x_b, x_b_s, x_r_s, x_c_s,
          sparsity_lut_x, s_lut_r, s_lut_r_s, s_lut_c_s,
          y, y_b, y_b_s, y_r_s, y_c_s,
          s_l_y_b, s_l_y_b_s, s_l_y_r_s,
          sparsity_reverse_x_lut_rwm,
          output,
          o_b, o_b_s, o_r_s, o_c_s,
          sparsity_block_size))

        return output


# noinspection PyUnusedLocal
@triton.autotune(
    configs=get_autotune_configs(),
    key=["sparsity_block_size"],
    prune_configs_by={"early_config_prune": prune_autotune_configs},
    reset_to_zero=["o"]
)
@triton.jit
def row_wise_add_kernel(x,
                        x_b, x_b_s, x_r_s, x_c_s,
                        s_lut_x, s_lut_x_r, s_lut_x_r_s, s_lut_x_c_s,
                        y, y_b, y_b_s, y_r_s, y_c_s,
                        s_l_y_b, s_l_y_b_s, s_l_y_r_s,
                        r_lut_y,
                        o,
                        o_b, o_b_s, o_r_s, o_c_s,
                        sparsity_block_size,
                        TRITON_BLOCK_SIZE: tl.constexpr) -> None:
    # Get triton block indices
    pid_blk = tl.program_id(axis=0)
    pid_row = tl.program_id(axis=1)
    pid_col = tl.program_id(axis=2)

    # Get position of current sparsity block consisting of its batch and row index
    spa_val_idx = pid_blk * s_lut_x_r_s + tl.arange(0, 4) * s_lut_x_c_s
    spa_val_msk = (tl.arange(0, 4) < 3)
    spa_val = tl.load(s_lut_x + spa_val_idx, mask=spa_val_msk)

    spa_bat_x = tl.sum(spa_val * (tl.arange(0, 4) == 0))
    spa_row_x = tl.sum(spa_val * (tl.arange(0, 4) == 1))
    spa_col_x = tl.sum(spa_val * (tl.arange(0, 4) == 2))

    # Get reverse sparsity indices for s
    rev_idx_spa_s_idx = (spa_bat_x * s_l_y_b_s +
                         spa_row_x * s_l_y_r_s)
    rev_idx_spa_s_msk = ((rev_idx_spa_s_idx >= 0) &
                         (rev_idx_spa_s_idx < s_l_y_b * s_l_y_b_s))
    rev_idx_spa_s = tl.load(r_lut_y + rev_idx_spa_s_idx, mask=rev_idx_spa_s_msk).to(tl.int32)

    if rev_idx_spa_s == -1:
        tl.device_assert(False)
        return

    # Load x block
    blk_x_idx = ((pid_blk * x_b_s) +
                 ((pid_row * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * x_r_s)[:, None] +
                 ((pid_col * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * x_c_s)[None, :])
    blk_x_msk = ((blk_x_idx >= 0) &
                 (blk_x_idx < x_b * x_b_s))
    blk_x = tl.load(x + blk_x_idx, mask=blk_x_msk)

    # Load sum block
    blk_s_idx = (rev_idx_spa_s * y_b_s +
                 ((pid_row * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * y_r_s)[:, None] +
                 (tl.arange(0, 1) * y_c_s)[None, :])
    blk_s_msk = ((blk_s_idx >= 0) &
                 (blk_s_idx < y_b * y_b_s))
    blk_s = tl.load(y + blk_s_idx, mask=blk_s_msk)

    # Compute exp
    buf = blk_x + tl.broadcast_to(blk_s, (TRITON_BLOCK_SIZE, TRITON_BLOCK_SIZE))

    # Store block
    blk_o_idx = ((pid_blk * o_b_s) +
                 ((pid_row * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * o_r_s)[:, None] +
                 ((pid_col * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * o_c_s)[None, :])
    blk_o_msk = ((blk_o_idx >= 0) &
                 (blk_o_idx < o_b * o_b_s))
    tl.store(o + blk_o_idx, buf, mask=blk_o_msk)



import torch
from torch import Tensor
from torch._library import triton_op

from blksprs.ops.flow import flow_pull_forward, flow_push_forward
from blksprs.utils.blksprs_tensor import BlksprsTensor
from blksprs.utils.validation import validate_dimensions, validate_contiguous, validate_device, \
    validate_sparsity, validate_sparsity_block_size, ensure_contiguous


@torch.amp.custom_fwd(device_type="cuda", cast_inputs=torch.float16)
def repeat(x: BlksprsTensor, sparsity_layout_x: Tensor, repeats: tuple[int, int, int],
           sparsity_block_size: int, sparsity_layout_output: Tensor = None, lut: dict = None) -> (
        BlksprsTensor, Tensor):
    """Repeats a block-spare tensor in compressed form according to the given repeats.
    
    Repeats is a 3-tuple of integers, where each integer represents the number of times the tensor should be repeated in
        the first, second and third dimension respectively.
        
    Note:
        An output sparsity layout can be provided, in which case only the indicated blocks are filled. This may result
        in blocks not being present in the output that were present in the input if the output sparsity layout indicates
        them to be sparse.
    
    Args:
        x (BlksprsTensor): A block-sparse tensor in compressed form.
        sparsity_layout_x (Tensor): The sparsity layout of the block-sparse tensor.
        repeats (tuple[int, int, int]): The number of times the tensor should be repeated in the first, second and
            third dimension respectively.
        sparsity_block_size (int): The size of the sparsity blocks.
        sparsity_layout_output (Tensor): The desired sparsity layout of the output tensor (default ``None``).
        lut (dict, optional): A dictionary containing the look-up tables for the operation (default ``None``).

    Returns:
        BlksprsTensor: A block-sparse tensor in compressed form containing the repeated values.
        Tensor: The sparsity layout of the resulting output tensor.

    """
    x = ensure_contiguous(x)

    validate_dimensions(x)
    validate_contiguous(x)
    validate_device(x)
    validate_sparsity(sparsity_block_size, (x, sparsity_layout_x))
    validate_sparsity_block_size(sparsity_block_size, x)

    lut = repeat_build_lut(lut, sparsity_layout_x, repeats, sparsity_layout_output)

    return BlksprsTensor.wrap(repeat_forward(
        x, sparsity_layout_x, lut["sparsity_layout_o"], lut["sparsity_lut"],
        lut["sparsity_reverse_lut"], sparsity_block_size, lut["n_sparse_blocks"])), lut["sparsity_layout_o"]


@torch.amp.custom_fwd(device_type="cuda", cast_inputs=torch.float16)
def repeat_interleave(x: BlksprsTensor, sparsity_layout_x: Tensor, repeats: int,
                      sparsity_block_size: int, sparsity_layout_output: Tensor = None, lut: dict = None) -> (
        BlksprsTensor, Tensor):
    """Repeats and interleaves the block-sparse tensor in compressed form.

    Repeats each matrix contained in the tensors by ``repeats`` amount and places them consecutively in the output
        tensor.

    Note:
        In similar fashion to the regular ``repeat`` an output sparsity layout can be provided. In this case only
        non-sparse blocks will be filled.

    Args:
        x (BlksprsTensor): A block-sparse tensor in compressed form.
        sparsity_layout_x (Tensor): The sparsity layout of the block-sparse tensor.
        repeats (int): The number of times to repeat the matrices.
        sparsity_block_size (int): The size of the sparsity blocks.
        sparsity_layout_output (Tensor): The desired sparsity layout of the output tensor (default ``None``).
        lut (dict, optional): A dictionary containing the look-up tables for the operation (default ``None``).

    Returns:
        BlksprsTensor: A block-sparse tensor in compressed form containing the repeated and interleaved matrices.
        Tensor: The sparsity layout of the resulting output tensor.

    """
    x = ensure_contiguous(x)

    validate_dimensions(x)
    validate_contiguous(x)
    validate_device(x)
    validate_sparsity(sparsity_block_size, (x, sparsity_layout_x))
    validate_sparsity_block_size(sparsity_block_size, x)

    lut = repeat_interleave_build_lut(lut, sparsity_layout_x, repeats, sparsity_layout_output)

    return BlksprsTensor.wrap(repeat_forward(
        x, sparsity_layout_x, lut["sparsity_layout_o"], lut["sparsity_lut"],
        lut["sparsity_reverse_lut"], sparsity_block_size, lut["n_sparse_blocks"])), lut["sparsity_layout_o"]


@triton_op("blksprs::repeat_forward", mutates_args={})
def repeat_forward(x: Tensor, _: Tensor, sparsity_layout_o: Tensor, sparsity_lut: Tensor,
                   sparsity_reverse_lut: Tensor,
                   sparsity_block_size: int, n_sparse_blocks: int) -> Tensor:
    with torch.no_grad():
        return flow_pull_forward(x, sparsity_layout_o, sparsity_lut, sparsity_reverse_lut, sparsity_block_size,
                                 n_sparse_blocks)


def repeat_wrapper_backward(ctx, grad_output):
    sparsity_layout_x, sparsity_layout_o, sparsity_lut, sparsity_reverse_lut = ctx.saved_tensors
    sparsity_block_size = ctx.sparsity_block_size
    n_sparse_blocks = torch.sum(sparsity_layout_x.to(torch.int)).item()

    return flow_push_forward(grad_output, sparsity_layout_o, sparsity_lut,
                             sparsity_reverse_lut, sparsity_block_size,
                             n_sparse_blocks), None, None, None, None, None, None


def repeat_build_lut(lut: dict, sparsity_layout_x: Tensor, repeats: tuple[int, int, int],
                     sparsity_layout_output: Tensor):
    if lut is None:
        lut = dict()

    if "sparsity_layout_o" not in lut:
        sparsity_layout_o = sparsity_layout_x.repeat(repeats[0], repeats[1], repeats[2])
        lut["sparsity_layout_o"] = sparsity_layout_o

    if sparsity_layout_output is not None:
        sparsity_layout_o = torch.logical_and(lut["sparsity_layout_o"], sparsity_layout_output)
        lut["sparsity_layout_o"] = sparsity_layout_o

    if "sparsity_lut" not in lut:
        sparsity_lut = torch.nonzero(lut["sparsity_layout_o"]).contiguous()
        lut["sparsity_lut"] = sparsity_lut

    if "sparsity_reverse_lut" not in lut:
        sparsity_layout_flat = sparsity_layout_x.reshape(-1)
        sparsity_reverse_lut = (((torch.cumsum(sparsity_layout_flat, dim=-1) - 1) *
                                 (sparsity_layout_flat == 1) -
                                 (1 * (sparsity_layout_flat == 0)))
                                .reshape(sparsity_layout_x.size())
                                .repeat(repeats[0], repeats[1], repeats[2])
                                .reshape(-1).contiguous())
        lut["sparsity_reverse_lut"] = sparsity_reverse_lut

    if "n_sparse_blocks" not in lut:
        n_sparse_blocks = torch.sum(lut["sparsity_layout_o"].to(torch.int)).item()
        lut["n_sparse_blocks"] = n_sparse_blocks

    validate_contiguous(lut["sparsity_layout_o"], lut["sparsity_lut"], lut["sparsity_reverse_lut"])

    return lut


def repeat_interleave_build_lut(lut: dict, sparsity_layout_x: Tensor, repeats: int,
                                sparsity_layout_output: Tensor):
    if lut is None:
        lut = dict()

    if "sparsity_layout_o" not in lut:
        sparsity_layout_o = torch.repeat_interleave(sparsity_layout_x, repeats, dim=0).contiguous()
        lut["sparsity_layout_o"] = sparsity_layout_o

    if sparsity_layout_output is not None:
        sparsity_layout_o = torch.logical_and(lut["sparsity_layout_o"], sparsity_layout_output)
        lut["sparsity_layout_o"] = sparsity_layout_o

    if "sparsity_lut" not in lut:
        sparsity_lut = torch.nonzero(lut["sparsity_layout_o"]).contiguous()
        lut["sparsity_lut"] = sparsity_lut

    if "sparsity_reverse_lut" not in lut:
        sparsity_layout_flat = sparsity_layout_x.reshape(-1)
        sparsity_reverse_lut = (((torch.cumsum(sparsity_layout_flat, dim=-1) - 1) *
                                 (sparsity_layout_flat == 1) -
                                 (1 * (sparsity_layout_flat == 0)))
                                .reshape(sparsity_layout_x.size())
                                .repeat_interleave(repeats, dim=0)
                                .reshape(-1).contiguous())
        lut["sparsity_reverse_lut"] = sparsity_reverse_lut

    if "n_sparse_blocks" not in lut:
        n_sparse_blocks = torch.sum(lut["sparsity_layout_o"].to(torch.int)).item()
        lut["n_sparse_blocks"] = n_sparse_blocks

    validate_contiguous(lut["sparsity_layout_o"], lut["sparsity_lut"], lut["sparsity_reverse_lut"])

    return lut


# noinspection PyUnusedLocal
def repeat_setup_context(ctx, inputs, output):
    (_, sparsity_layout_x, sparsity_layout_o, sparsity_lut, sparsity_reverse_lut, sparsity_block_size, _) = inputs

    ctx.save_for_backward(sparsity_layout_x, sparsity_layout_o, sparsity_lut, sparsity_reverse_lut)
    ctx.sparsity_block_size = sparsity_block_size


repeat_forward.register_autograd(repeat_wrapper_backward, setup_context=repeat_setup_context)



from typing import Union

import torch
from torch import Tensor


class BlksprsTensor(Tensor):
    """A wrapper class representing a block-sparse tensor in compressed form.
    """

    def __repr__(self):
        return f"BlksprsTensor({torch.Tensor(self).__repr__()})"

    @staticmethod
    def wrap(tensor: Tensor) -> Union[Tensor, "BlksprsTensor"]:
        if torch._dynamo.is_compiling():
            return tensor
        else:
            return BlksprsTensor(tensor)



import math

import torch
import triton
from torch import Tensor
from torch._library.triton import wrap_triton, triton_op
from triton import language as tl

from blksprs.utils.autotuning import get_autotune_configs, prune_autotune_configs, prune_autotune_configs_conversion
from blksprs.utils.blksprs_tensor import BlksprsTensor
from blksprs.utils.tools import stride
from blksprs.utils.validation import validate_dimensions, validate_device, \
    validate_contiguous, validate_sparsity, validate_sparsity_block_size


@torch.amp.custom_fwd(device_type="cuda", cast_inputs=torch.float16)
def build_sparsity_layout(x: Tensor, sparsity_block_size: int) -> Tensor:
    """Builds the sparsity layout of a dense tensor in regular form covering its sparse blocks.

    Args:
        x (Tensor): A block-sparse (or dense) tensor in regular form.
        sparsity_block_size (int): The size of the sparsity blocks.

    Returns:
        Tensor: The sparsity layout of the input block-sparse (or dense) tensor.

    """
    validate_dimensions(x)
    validate_contiguous(x)
    validate_device(x)

    return build_sparsity_layout_operation(x, sparsity_block_size)


@triton_op("blksprs::build_sparsity_layout", mutates_args={})
def build_sparsity_layout_operation(x: Tensor, sparsity_block_size: int) -> Tensor:
    with torch.no_grad():
        output = torch.zeros(x.size(0), x.size(1) // sparsity_block_size, x.size(2) // sparsity_block_size,
                             dtype=torch.bool, device=x.device)

        x_b, x_r, x_c = x.size()
        x_b_s, x_r_s, x_c_s = stride(x)
        o_b, o_r, o_c = output.size()
        o_b_s, o_r_s, o_c_s = stride(output)

        triton_grid = lambda meta: [x_b,
                                    triton.cdiv(x_r, meta["TRITON_BLOCK_SIZE"]),
                                    triton.cdiv(x_c, meta["TRITON_BLOCK_SIZE"])]

        (wrap_triton(build_sparsity_layout_kernel)[triton_grid]
         (x,
          x_b, x_b_s, x_r_s, x_c_s,
          output,
          o_b, o_b_s, o_r_s, o_c_s,
          sparsity_block_size))

        return output


@triton.autotune(
    configs=get_autotune_configs(),
    key=["sparsity_block_size"],
    prune_configs_by={"early_config_prune": prune_autotune_configs},
    reset_to_zero=["o"]
)
@triton.jit
def build_sparsity_layout_kernel(x,
                                 x_b, x_b_s, x_r_s, x_c_s,
                                 o,
                                 o_b, o_b_s, o_r_s, o_c_s,
                                 sparsity_block_size,
                                 TRITON_BLOCK_SIZE: tl.constexpr) -> None:
    # Get triton block indices
    pid_bat = tl.program_id(axis=0)
    pid_row = tl.program_id(axis=1)
    pid_col = tl.program_id(axis=2)

    # Load x values
    blk_x_idx = (pid_bat * x_b_s +
                 ((pid_row * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * x_r_s)[:, None] +
                 ((pid_col * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * x_c_s)[None, :])
    blk_x_msk = ((blk_x_idx >= 0) &
                 (blk_x_idx < x_b * x_b_s))
    blk_x = tl.load(x + blk_x_idx, mask=blk_x_msk)

    # Store sparsity layout value
    if tl.min(blk_x) != 0 or tl.max(blk_x) != 0:
        blk_o_idx = (pid_bat * o_b_s +
                     (((pid_row * TRITON_BLOCK_SIZE) // sparsity_block_size) * o_r_s +
                      ((pid_col * TRITON_BLOCK_SIZE) // sparsity_block_size) * o_c_s))
        blk_o_msk = ((blk_o_idx >= 0) &
                     (blk_o_idx < o_b * o_b_s))
        tl.store(o + blk_o_idx, 1, mask=blk_o_msk)


@torch.amp.custom_fwd(device_type="cuda", cast_inputs=torch.float16)
def build_sparsity_layout_adaption(x: BlksprsTensor, sparsity_layout_from: Tensor,
                                   sparsity_block_size_from: int, sparsity_block_size_to: int) -> Tensor:
    """Builds the sparsity layout of a block-sparse tensor in compressed form if a different sparsity block size were
        used.
        
    Args:
        x (BlksprsTensor): A block-sparse tensor in compressed form.
        sparsity_layout_from (Tensor): The sparsity layout of the input block-sparse tensor.
        sparsity_block_size_from (int): The size of the sparsity blocks of the input tensor.
        sparsity_block_size_to (int): The desired size of the sparsity blocks for the resulting layout.

    Returns:
        Tensor: The sparsity layout in regular form using the new sparsity block size of the input block-sparse tensor
            in compressed form.
    
    """
    validate_dimensions(x)
    validate_contiguous(x, sparsity_layout_from)
    validate_device(x)
    validate_sparsity(sparsity_block_size_from, (x, sparsity_layout_from))
    validate_sparsity_block_size(sparsity_block_size_from, x)
    validate_sparsity_block_size(sparsity_block_size_to)

    sparsity_lut = torch.nonzero(sparsity_layout_from).contiguous()

    validate_contiguous(sparsity_layout_from, sparsity_lut)

    return build_sparsity_layout_adaption_operation(x, sparsity_layout_from, sparsity_lut,
                                                    sparsity_block_size_from, sparsity_block_size_to)


@triton_op("blksprs::build_sparsity_layout_adaption", mutates_args={})
def build_sparsity_layout_adaption_operation(x: Tensor, sparsity_layout_from: Tensor, sparsity_lut: Tensor,
                                             sparsity_block_size_from: int, sparsity_block_size_to: int) -> Tensor:
    with torch.no_grad():
        o_b = sparsity_layout_from.size(0)
        o_r = math.ceil(sparsity_layout_from.size(1) * sparsity_block_size_from // sparsity_block_size_to)
        o_c = math.ceil(sparsity_layout_from.size(2) * sparsity_block_size_from // sparsity_block_size_to)

        output = torch.zeros(o_b, o_r, o_c, dtype=torch.bool, device=x.device)

        x_b, x_r, x_c = x.size()
        x_b_s, x_r_s, x_c_s = stride(x)
        s_lut_r, s_lut_c = sparsity_lut.size()
        s_lut_r_s, s_lut_c_s = stride(sparsity_lut)
        o_b_s, o_r_s, o_c_s = stride(output)

        triton_grid = lambda meta: [x_b,
                                    triton.cdiv(x_r, meta["TRITON_BLOCK_SIZE"]),
                                    triton.cdiv(x_c, meta["TRITON_BLOCK_SIZE"])]

        (wrap_triton(build_sparsity_layout_adaption_kernel)[triton_grid]
         (x,
          x_b, x_b_s, x_r_s, x_c_s,
          sparsity_lut, s_lut_r, s_lut_r_s, s_lut_c_s,
          output,
          o_b, o_b_s, o_r_s, o_c_s,
          sparsity_block_size_from,
          sparsity_block_size_to))

        return output


@triton.autotune(
    configs=get_autotune_configs(),
    key=["sparsity_block_size_from", "sparsity_block_size_to"],
    prune_configs_by={"early_config_prune": prune_autotune_configs_conversion},
    reset_to_zero=["o"]
)
@triton.jit
def build_sparsity_layout_adaption_kernel(x,
                                          x_b, x_b_s, x_r_s, x_c_s,
                                          s_lut, s_lut_r, s_lut_r_s, s_lut_c_s,
                                          o,
                                          o_b, o_b_s, o_r_s, o_c_s,
                                          sparsity_block_size_from,
                                          sparsity_block_size_to,
                                          TRITON_BLOCK_SIZE: tl.constexpr) -> None:
    # Get triton block indices
    pid_blk = tl.program_id(axis=0)
    pid_row = tl.program_id(axis=1)
    pid_col = tl.program_id(axis=2)

    # Get sparsity index of current output block consisting of its batch, row, and column index
    spa_bat_idx = (pid_blk * s_lut_r_s + 0 * s_lut_c_s)
    spa_bat_msk = ((spa_bat_idx >= 0) &
                   (spa_bat_idx < s_lut_r * s_lut_r_s))
    spa_bat = tl.load(s_lut + spa_bat_idx, mask=spa_bat_msk)

    spa_row_idx = (pid_blk * s_lut_r_s + 1 * s_lut_c_s)
    spa_row_msk = ((spa_row_idx >= 0) &
                   (spa_row_idx < s_lut_r * s_lut_r_s))
    spa_row = tl.load(s_lut + spa_row_idx, mask=spa_row_msk)

    spa_col_idx = (pid_blk * s_lut_r_s + 2 * s_lut_c_s)
    spa_col_msk = ((spa_col_idx >= 0) &
                   (spa_col_idx < s_lut_r * s_lut_r_s))
    spa_col = tl.load(s_lut + spa_col_idx, mask=spa_col_msk)

    # Load x values
    blk_x_idx = ((pid_blk * x_b_s) +
                 ((pid_row * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * x_r_s)[:, None] +
                 ((pid_col * TRITON_BLOCK_SIZE + tl.arange(0, TRITON_BLOCK_SIZE)) * x_c_s)[None, :])
    blk_x_msk = ((blk_x_idx >= 0) &
                 (blk_x_idx < x_b * x_b_s))
    blk_x = tl.load(x + blk_x_idx, mask=blk_x_msk)

    # Store sparsity layout value
    if tl.min(blk_x) != 0 or tl.max(blk_x) != 0:
        blk_o_idx = ((spa_bat * o_b_s) +
                     (((pid_row * TRITON_BLOCK_SIZE + spa_row * sparsity_block_size_from)
                       // sparsity_block_size_to) * o_r_s) +
                     (((pid_col * TRITON_BLOCK_SIZE + spa_col * sparsity_block_size_from)
                       // sparsity_block_size_to) * o_c_s))
        blk_o_msk = ((blk_o_idx >= 0) &
                     (blk_o_idx < o_b * o_b_s))
        tl.store(o + blk_o_idx, 1, mask=blk_o_msk)


@torch.amp.custom_fwd(device_type="cuda", cast_inputs=torch.float16)
def build_sparsity_layout_matmul(sparsity_layout_x: Tensor, sparsity_layout_y: Tensor) -> Tensor:
    """Builds the precise sparsity layout of the result of a matrix multiplication between the two input tensors.

    Args:
        sparsity_layout_x (Tensor): The sparsity layout of the first block-sparse tensor.
        sparsity_layout_y (Tensor): The sparsity layout of the second block-sparse tensor.

    Returns:
        Tensor: The precise sparsity layout of the result of a matrix multiplication between the two input tensors.

    """
    return torch.matmul(sparsity_layout_x.to(torch.float), sparsity_layout_y.to(torch.float)).to(torch.bool)


@torch.amp.custom_fwd(device_type="cuda", cast_inputs=torch.float16)
def build_sparsity_layout_matmul_fast(sparsity_layout_x: Tensor, sparsity_layout_y: Tensor):
    """Builds the approximate sparsity layout of the result of a matrix multiplication between the two input tensors.

    Note:
        This function is faster than the ``build_sparsity_layout_matmul`` function due to the fact that it only checks
            whether at least one of the blocks in either of the vectors participating in the matmul is non-sparse. The
            resulting sparsity layout may thus overestimate the actual sparsity of the result.

    Args:
        sparsity_layout_x (Tensor): The sparsity layout of the first block-sparse tensor.
        sparsity_layout_y (Tensor): The sparsity layout of the second block-sparse tensor.

    Returns:
        Tensor: The approximate sparsity layout of the result of a matrix multiplication between the two input tensors.

    """
    sparsity_layout_x_slice = torch.max(sparsity_layout_x, dim=-1).values.unsqueeze(-1)
    sparsity_layout_y_slice = torch.max(sparsity_layout_y, dim=-2).values.unsqueeze(1)

    return torch.logical_or(sparsity_layout_x_slice, sparsity_layout_y_slice)


def build_sparsity_layout_full(x: Tensor, sparsity_block_size: int) -> Tensor:
    return torch.ones(size=(x.size(0), x.size(1) // sparsity_block_size, x.size(2) // sparsity_block_size),
                      dtype=torch.bool, device=x.device)



from typing import Callable

import triton
from matplotlib import pyplot as plt


def benchmark(method_labels: list[str], func_input_generator: Callable,
              matrix_sizes: list[int], sparsity_block_sizes: list[int],
              *funcs_test_subject: Callable, y_lim_top: int = None):
    quantiles = [0.5, 0.2, 0.8]
    results = {}

    for matrix_size, sparsity_block_size in zip(matrix_sizes, sparsity_block_sizes):
        arguments = func_input_generator(matrix_size, sparsity_block_size)

        for i, func_test_subject in enumerate(funcs_test_subject):
            func_ms_avg, func_ms_min, func_ms_max = triton.testing.do_bench(
                lambda: func_test_subject(**arguments), quantiles=quantiles)
            results.setdefault(i, []).append((func_ms_avg, func_ms_min, func_ms_max))

    plt.figure(dpi=300)
    for key_method, value_method in results.items():
        ms_method_avg, ms_method_min, ms_method_max = zip(*value_method)
        plt.plot(matrix_sizes, ms_method_avg, label=method_labels[key_method])
        plt.fill_between(matrix_sizes, ms_method_min, ms_method_max, alpha=0.2)

    plt.xlabel("Matrix size")
    plt.ylabel("Time (ms)")
    plt.ylim(bottom=0, top=y_lim_top)
    plt.grid(True)
    plt.legend()
    plt.show()



from torch import Tensor, Size


def do_shape_blocksparse(x: Tensor) -> tuple[Tensor, Size]:
    if x.dim() == 3:
        return x.contiguous(), x.size()

    return x.reshape(-1, x.size(-2), x.size(-1)).contiguous(), x.size()


def undo_shape_blocksparse(x: Tensor, shape: Size | tuple[int, ...]) -> Tensor:
    if x.shape[:-2] == shape[:-2]:
        return x

    return x.reshape((*shape[:-2], *x.shape[-2:]))


def stride(x: Tensor):
    if x.dim() == 1:
        return 1
    elif x.dim() == 2:
        return x.size(1), 1
    elif x.dim() == 3:
        return x.size(1) * x.size(2), x.size(2), 1
    else:
        raise NotImplementedError


def ceil_pow2(x: int) -> int:
    if x <= 0:
        raise ValueError("Input must be a positive integer.")
    return 1 << (x - 1).bit_length()



from collections.abc import Callable

import torch
from torch import Tensor, nn

import blksprs as bs
from blksprs.layouting.sparsity_layout import build_sparsity_layout_matmul_fast
from blksprs.ops.conversion import to_sparse
from blksprs.ops.matmul import matmul
from blksprs.ops.repeat import repeat
from blksprs.utils.blksprs_tensor import BlksprsTensor


@torch.amp.custom_fwd(device_type="cuda", cast_inputs=torch.float16)
def apply_torch_linear(x: BlksprsTensor, sparsity_layout: Tensor, sparsity_block_size: int,
                       linear: nn.Linear, bias: nn.Parameter = None) -> (BlksprsTensor, Tensor):
    # Extract weight and bias
    w = linear.weight
    b = linear.bias

    # Convert w to block-sparse representation
    sparsity_layout_w_t = torch.ones(size=(sparsity_layout.size(0), w.size(1) // sparsity_block_size,
                                           w.size(0) // sparsity_block_size), dtype=torch.bool, device=x.device)
    w_t_bs = to_sparse(w.transpose(-1, -2).unsqueeze(0).repeat(sparsity_layout.size(0), 1, 1),
                       sparsity_layout_w_t, sparsity_block_size)

    # Apply weights
    sparsity_layout_xw = build_sparsity_layout_matmul_fast(sparsity_layout, sparsity_layout_w_t)
    xw = matmul(x, sparsity_layout, BlksprsTensor.wrap(w_t_bs.to(x.dtype)), sparsity_layout_w_t, sparsity_layout_xw,
                sparsity_block_size)
    interim = xw

    # Apply bias
    if bias is not None:
        b = bias
    if b is not None:
        b_slice = b.unsqueeze(0).unsqueeze(0).repeat(1, sparsity_block_size, 1)
        sparsity_layout_b_slice = torch.ones(size=(1, b_slice.size(1) // sparsity_block_size,
                                                   b_slice.size(2) // sparsity_block_size), dtype=torch.bool,
                                             device=x.device)
        b_slice_bs = to_sparse(b_slice, sparsity_layout_b_slice, sparsity_block_size)
        b_bs, sparsity_layout_b = repeat(b_slice_bs, sparsity_layout_b_slice,
                                         (sparsity_layout.size(0), sparsity_layout_xw.size(1), 1), sparsity_block_size,
                                         sparsity_layout_output=sparsity_layout_xw)
        interim = interim + b_bs

    return interim, sparsity_layout_xw


def apply_torch_normalisation(x: BlksprsTensor, sparsity_layout: Tensor, sparsity_block_size: int,
                              normalisation: nn.Module) -> BlksprsTensor:
    return apply_function_applicable_row_wise(x, sparsity_layout, sparsity_block_size, normalisation)


def apply_torch_dropout(x: BlksprsTensor, sparsity_layout: Tensor, sparsity_block_size: int,
                        dropout: nn.Dropout) -> BlksprsTensor:
    return apply_function_applicable_row_wise(x, sparsity_layout, sparsity_block_size, dropout)


def apply_function_applicable_row_wise(x: BlksprsTensor, sparsity_layout: Tensor, sparsity_block_size: int,
                                       function: Callable) -> BlksprsTensor:
    sparsity_layout_packed = _pack_layout(sparsity_layout)
    blksprs_pseudo_dense = bs.ops.to_dense(x, sparsity_layout_packed, sparsity_block_size)
    normalisation_out = function(blksprs_pseudo_dense)
    blksprs_sparse = bs.ops.to_sparse(normalisation_out, sparsity_layout_packed, sparsity_block_size)

    return blksprs_sparse


def _pack_layout(sparsity_layout: Tensor) -> BlksprsTensor:
    sparsity_layout_reshaped = sparsity_layout.reshape(1, sparsity_layout.size(0) * sparsity_layout.size(1),
                                                       sparsity_layout.size(2))
    non_zero_rows = torch.any(sparsity_layout_reshaped, dim=-1)
    sparsity_layout_filtered = sparsity_layout_reshaped[non_zero_rows].unsqueeze(0)

    return sparsity_layout_filtered
